# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from kubemarine import kubernetes
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup


def k8s_certs_overview(control_planes: NodeGroup) -> None:
    for control_plane in control_planes.get_ordered_members_list():
        control_planes.cluster.log.debug(f"Checking certs expiration for control_plane {control_plane.get_node_name()}")
        control_plane.sudo("kubeadm certs check-expiration", hide=False)


def renew_verify(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get('initial_procedure') != 'cert_renew' or "kubernetes" not in cluster.procedure_inventory:
        return inventory

    cert_list = cluster.procedure_inventory["kubernetes"].get("cert-list")
    verify_all_is_absent_or_single(cert_list)

    return inventory


def renew_apply(control_planes: NodeGroup) -> None:
    log = control_planes.cluster.log

    procedure = control_planes.cluster.procedure_inventory["kubernetes"]
    cert_list = procedure["cert-list"]

    for cert in cert_list:
        control_planes.sudo(f"kubeadm certs renew {cert}")

    if "all" in cert_list or "admin.conf" in cert_list:
        # need to update cluster-admin config
        kubernetes.copy_admin_config(log, control_planes)

    # for some reason simple pod delete do not work for certs update - we need to delete containers themselves
    control_planes.call(force_restart_control_plane)

    for control_plane in control_planes.get_ordered_members_list():
        kubernetes.wait_for_any_pods(control_planes.cluster, control_plane, apply_filter=control_plane.get_node_name())


def force_restart_control_plane(control_planes: NodeGroup) -> None:
    cri_impl = control_planes.cluster.inventory['services']['cri']['containerRuntime']
    restart_containers = ["etcd", "kube-scheduler", "kube-apiserver", "kube-controller-manager"]
    c_filter = "grep -e %s" % " -e ".join(restart_containers)

    if cri_impl == "docker":
        control_planes.sudo("sudo docker container rm -f $(sudo docker ps -a | %s | awk '{ print $1 }')" % c_filter, warn=True)
    else:
        control_planes.sudo("sudo crictl rm -f $(sudo crictl ps -a | %s | awk '{ print $1 }')" % c_filter, warn=True)


def verify_all_is_absent_or_single(cert_list: List[str]) -> None:
    if "all" in cert_list and len(cert_list) > 1:
        raise Exception(f"Found 'all' in certs list, but it is not single: {cert_list}")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import sys
import traceback

from concurrent.futures import TimeoutError
from typing import Type

from kubemarine.core import log as klog


def get_kme_dictionary() -> dict:
    from kubemarine.core.group import GroupException
    return {
        "KME0000": {
            "name": "Test exception"
        },
        "KME0002": {
            "instance": GroupException,
            "name": "Remote group exception\n{reason}"
        },
        "KME0003": {
            "instance": TimeoutError,
            "name": "Action took too long to complete and timed out"
        },
        "KME0004": {
            "name": "There are no workers defined in the cluster scheme"
        },
        "KME0005": {
            "name": "{hostnames} are not sudoers"
        },
        "KME0007": {
            "name": "Docker CRI can not be used with endpoints registry definition."
        },
        "KME0008": {
            "name": "Specified Kubernetes version '{version}' - cannot be used! "
                    "Allowed versions are: {allowed_versions}."
        },
        "KME0009": {
            "name": "Key {key!r} is redefined for {plugin_name!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required plugin configuration explicitly in procedure inventory."
        },
        "KME0010": {
            "name": "Associations are redefined for {package!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required associations explicitly in procedure inventory."
        },
        "KME0011": {
            "name": "Key {key!r} is redefined for third-party {thirdparty!r} in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required third-party configuration explicitly in procedure inventory."
        },
        "KME0012": {
            "name": "Procedure {procedure!r} is possible only for cluster "
                    "with all nodes having the same and supported OS family."
        },
        'KME0013': {
            "name": "Key 'plugins.\"io.containerd.grpc.v1.cri\".sandbox_image' is redefined for 'containerdConfig' "
                    "in cluster.yaml{previous_version_spec}, "
                    "but not present in procedure inventory{next_version_spec}. "
                    "Please, specify required 'sandbox_image' explicitly in procedure inventory."
        }
    }


# TODO: support for more complex KME00XX objects with custom constructors
class KME(RuntimeError):
    def __init__(self, code: str, **kwargs: object):
        self.code = code
        self.kme = get_kme_dictionary().get(self.code)
        if self.kme is None:
            raise ValueError('An error was raised with an unknown error code')
        name: str = self.kme.get('name')
        self.message = name.format_map(kwargs)
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.code + ": " + self.message


class FailException(Exception):
    def __init__(self, message: str = '', reason: Exception = None, hint: str = '') -> None:
        super().__init__(message)
        self.message = message
        self.reason = reason
        self.hint = hint


def wrap_kme_exception(reason: Exception) -> Exception:
    for dictionary_code, dictionary_kme in get_kme_dictionary().items():
        exception_class: Type[Exception] = dictionary_kme.get('instance')
        if exception_class is not None and isinstance(reason, exception_class):
            return KME(dictionary_code, reason=reason)

    return reason


def pretty_print_error(message: str = '', reason: Exception = None, log: klog.EnhancedLogger = None) -> None:
    """
    Parses the passed error and nicely displays its name and structure depending on what was passed.
    The method outputs to stdout by default, but will use the logger if one is specified.

    :param message: an optional message describing context of the error
    :param reason: an exception that caused the failure.
    :param log: logger object, if you need to write a log there
    :return: None
    """
    def error_logger(msg: object) -> None:
        if log:
            log.critical(msg)
        else:
            sys.stderr.write(str(msg) + "\n")

    error_logger('FAILURE!')
    if message != '':
        error_logger(message)

    if reason is None:
        return

    reason = wrap_kme_exception(reason)

    if isinstance(reason, KME):
        error_logger(reason)
        return

    if log:
        log.critical('KME0001: Unexpected exception', exc_info=reason)
    else:
        sys.stderr.write("KME0001: Unexpected exception\n")
        exc_info = (type(reason), reason, reason.__traceback__)
        sio = io.StringIO()
        traceback.print_exception(*exc_info, limit=None, file=sio)
        sys.stderr.write(sio.getvalue())



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List, Dict, Tuple, Optional, Union, Mapping, Set

from typing_extensions import Protocol

from kubemarine import yum, apt
from kubemarine.core import errors, utils, static
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.executor import RunnersResult, Token, Callback
from kubemarine.core.group import (
    NodeGroup, DeferredGroup, AbstractGroup, RunResult, GROUP_RUN_TYPE,
    CollectorCallback, RunnersGroupResult
)
from kubemarine.core.yaml_merger import default_merger


ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS = \
    "It is not supported to customize services.packages.associations section " \
    "if nodes have different OS families. " \
    "Please move the section to corresponding services.packages.associations.<os_family> section."

ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED = \
    "Multiple package versions detected %s for package '%s'. " \
    "Align them to the single version manually or using corresponding task of install procedure. " \
    "Alternatively, specify cache_versions=false for corresponding association."

ERROR_SEMANAGE_NOT_MANAGED_DEBIAN = "semanage is not managed for debian OS family by KubeMarine"


def enrich_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    enrich_inventory_associations(inventory, cluster)
    enrich_inventory_packages(inventory, cluster)

    return inventory


def enrich_inventory_associations(inventory: dict, cluster: KubernetesCluster) -> None:
    associations: dict = inventory['services']['packages']['associations']
    enriched_associations = {}

    # Move associations for OS families and merge with globals
    for association_name in get_associations_os_family_keys():
        os_associations: dict = deepcopy(cluster.globals['packages']['common_associations'])
        if association_name == 'debian':
            del os_associations['semanage']
        for association_params in os_associations.values():
            del association_params['groups']
        default_merger.merge(os_associations, associations.pop(association_name))
        enriched_associations[association_name] = os_associations

    # Check remained associations section if they are customized at global level.
    if associations:
        os_family = cluster.get_os_family()
        if os_family == 'multiple':
            raise Exception(ERROR_GLOBAL_ASSOCIATIONS_REDEFINED_MULTIPLE_OS)
        elif os_family not in ('unknown', 'unsupported'):
            # move remained associations properties to the specific OS family section and merge with priority
            default_merger.merge(enriched_associations[os_family], associations)

    if 'semanage' in enriched_associations['debian']:
        raise Exception(ERROR_SEMANAGE_NOT_MANAGED_DEBIAN)

    inventory['services']['packages']['associations'] = enriched_associations


def enrich_inventory_packages(inventory: dict, _: KubernetesCluster) -> None:
    for _type in ['install', 'upgrade', 'remove']:
        packages_list = inventory['services']['packages'].get(_type)
        if isinstance(packages_list, list):
            inventory['services']['packages'][_type] = {
                'include': packages_list
            }


def enrich_inventory_apply_defaults(inventory: dict, _: KubernetesCluster) -> dict:
    kubernetes_version = inventory['services']['kubeadm']['kubernetesVersion']

    for os_family in get_associations_os_family_keys():
        cluster_associations = inventory["services"]["packages"]["associations"][os_family]
        for package in static.GLOBALS['packages'][os_family]:
            if cluster_associations[package].get('package_name') is None:
                cluster_associations[package]['package_name'] = \
                    get_default_package_names(os_family, package, kubernetes_version)

    return inventory


def _get_associations_upgrade_plan(cluster: KubernetesCluster, inventory: dict) -> List[Tuple[str, dict]]:
    context = cluster.context
    if context.get("initial_procedure") == "upgrade":
        upgrade_version = context["upgrade_version"]
        upgrade_plan = []
        for version in cluster.procedure_inventory['upgrade_plan']:
            if utils.version_key(version) < utils.version_key(upgrade_version):
                continue

            upgrade_associations = cluster.procedure_inventory.get(version, {}).get("packages", {}).get("associations", {})
            upgrade_associations = dict(item for item in upgrade_associations.items()
                                        if item[0] in _get_system_packages_support_upgrade(inventory))
            upgrade_plan.append((version, upgrade_associations))

    elif context.get("initial_procedure") == "migrate_kubemarine" and "upgrading_package" in context:
        upgrade_associations = cluster.procedure_inventory.get('upgrade', {}).get("packages", {}).get("associations", {})
        upgrade_associations = dict(item for item in upgrade_associations.items()
                                    if item[0] == context["upgrading_package"])
        upgrade_plan = [("", upgrade_associations)]
    else:
        upgrade_plan = []

    return upgrade_plan


def enrich_upgrade_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    upgrade_plan = _get_associations_upgrade_plan(cluster, inventory)
    if not upgrade_plan:
        return inventory

    os_family = cluster.get_os_family()
    if os_family not in get_associations_os_family_keys():
        raise errors.KME("KME0012", procedure='upgrade')

    context = cluster.context
    if context.get("initial_procedure") == "upgrade":
        previous_version = context["initial_kubernetes_version"]
        packages_verify = _get_system_packages_support_upgrade(inventory)
    else:  # migrate_kubemarine procedure
        previous_version = ""
        packages_verify = [context["upgrading_package"]]

    cluster_associations = inventory["services"]["packages"]["associations"][os_family]
    _verify_upgrade_plan(cluster_associations, previous_version, packages_verify, upgrade_plan)

    upgrade_required = get_system_packages_for_upgrade(cluster, inventory)
    context.setdefault("upgrade", {}).setdefault('required', {})['packages'] = upgrade_required

    # Merge procedure associations with the OS family specific section of associations in the inventory.
    upgrade_inventory_associations(cluster, inventory, enrich_global=False)
    upgrade_inventory_packages(cluster, inventory)

    return inventory


def upgrade_inventory_associations(cluster: KubernetesCluster, inventory: dict,
                                   *, enrich_global: bool) -> None:
    # pass enriched 'cluster.inventory' instead of 'inventory' that is being finalized
    upgrade_plan = _get_associations_upgrade_plan(cluster, cluster.inventory)
    if not upgrade_plan:
        return

    _, upgrade_associations = upgrade_plan[0]
    _enrich_inventory_procedure_associations(cluster, inventory, upgrade_associations,
                                             enrich_global=enrich_global)


def upgrade_inventory_packages(cluster: KubernetesCluster, inventory: dict) -> None:
    if cluster.context.get("initial_procedure") != "upgrade":
        return

    upgrade_version = cluster.context["upgrade_version"]
    for _type in ['install', 'upgrade', 'remove']:
        packages_section = cluster.procedure_inventory.get(upgrade_version, {}).get("packages", {})
        upgrade_packages = packages_section.get(_type)
        if upgrade_packages is None:
            continue
        if isinstance(upgrade_packages, list):
            upgrade_packages = {'include': upgrade_packages}

        packages_section = inventory.setdefault("services", {}).setdefault("packages", {})
        inventory_packages = packages_section.setdefault(_type, {})
        if isinstance(inventory_packages, list):
            packages_section[_type] = inventory_packages = {
                'include': inventory_packages
            }

        default_merger.merge(inventory_packages, upgrade_packages)


def _verify_upgrade_plan(cluster_associations: dict, previous_version: str,
                         packages_verify: List[str], upgrade_plan: List[Tuple[str, dict]]) -> None:
    cluster_associations = deepcopy(cluster_associations)

    # validate all packages sections in procedure inventory
    for version, upgrade_associations in upgrade_plan:
        for package in packages_verify:
            upgrade_package_name = upgrade_associations.get(package, {}).get('package_name')
            # Here default package names are not yet enriched and thus hold the custom supplied package names.
            if cluster_associations[package].get('package_name') and not upgrade_package_name:
                raise errors.KME("KME0010", package=package,
                                 previous_version_spec=f' for version {previous_version}' if previous_version else "",
                                 next_version_spec=f' for version {version}' if version else "")
            if upgrade_package_name:
                cluster_associations[package]['package_name'] = upgrade_package_name
        previous_version = version


def get_default_package_names(os_family: str, package: str, kubernetes_version: str) -> List[str]:
    version_key = get_compatibility_version_key(os_family)
    compatibility = static.GLOBALS['compatibility_map']['software']
    packages_names: List[Dict[str, str]] = static.GLOBALS['packages'][os_family][package]['package_name']

    package_versions = []
    for kv in packages_names:
        package_name, software_name = next((k, v) for k, v in kv.items())
        if software_name in ('haproxy', 'keepalived'):
            version = compatibility[software_name][version_key]
        else:
            version = compatibility[software_name][kubernetes_version][version_key]

        package_versions.append(f"{package_name}{get_package_version_separator(os_family)}{version}")

    return package_versions


def get_system_packages_for_upgrade(cluster: KubernetesCluster, inventory: dict) -> List[str]:
    undefined_package_name = object()

    context = cluster.context
    os_family = cluster.get_os_family()

    cluster_associations = inventory['services']['packages']['associations'][os_family]
    _, upgrade_associations = _get_associations_upgrade_plan(cluster, inventory)[0]

    # Resolve old and new packages with versions and schedule for upgrade if they are not equal.
    system_packages = _get_system_packages_support_upgrade(inventory)
    upgrade_required = []
    for package in system_packages:
        # Here default package names are not yet enriched and thus hold the custom supplied package names.
        old_package_name = cluster_associations[package].get('package_name')
        if old_package_name is None:
            # Trying enrichment before upgrade
            if context.get("initial_procedure") == "migrate_kubemarine":
                # Recommended versions have changed, and we forgot about what versions were previously recommended.
                old_package_name = undefined_package_name
            else:
                # upgrade procedure
                previous_ver = context["initial_kubernetes_version"]
                old_package_name = get_default_package_names(os_family, package, previous_ver)

        new_package_name = cluster_associations[package].get('package_name')
        # associations from procedure inventory have priority
        upgrade_package_name = upgrade_associations.get(package, {}).get('package_name')
        if upgrade_package_name:
            new_package_name = upgrade_package_name
        if new_package_name is None:
            # For upgrade procedure, services.kubeadm.kubernetesVersion is equal to 'upgrade_version',
            # because the version was already enriched.
            upgrade_ver = inventory['services']['kubeadm']['kubernetesVersion']
            new_package_name = get_default_package_names(os_family, package, upgrade_ver)

        if old_package_name is undefined_package_name or old_package_name != new_package_name:
            upgrade_required.append(package)

    return upgrade_required


def _get_system_packages_support_upgrade(inventory: dict) -> List[str]:
    return [inventory['services']['cri']['containerRuntime'], 'haproxy', 'keepalived']


def enrich_migrate_cri_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get("initial_procedure") != "migrate_cri":
        return inventory

    os_family = cluster.get_os_family()
    if os_family not in get_associations_os_family_keys():
        raise errors.KME("KME0012", procedure='migrate_cri')

    procedure_associations = cluster.procedure_inventory.get("packages", {}).get("associations", {})
    # Merge OS family specific section. It is already enriched in enrich_inventory_associations()
    # This effectively allows to specify only global section but not for specific OS family.
    # This restriction is because enrich_migrate_cri_inventory() goes after enrich_inventory_associations(),
    # but in future the restriction can be eliminated.
    return _enrich_inventory_procedure_associations(cluster, inventory, procedure_associations,
                                                    enrich_global=False)


def _enrich_inventory_procedure_associations(cluster: KubernetesCluster, inventory: dict,
                                             procedure_associations: dict,
                                             *, enrich_global: bool) -> dict:
    if procedure_associations:
        cluster_associations = inventory.setdefault("services", {}).setdefault("packages", {}) \
            .setdefault("associations", {})
        if not enrich_global:
            cluster_associations = cluster_associations.setdefault(cluster.get_os_family(), {})
        default_merger.merge(cluster_associations, deepcopy(procedure_associations))

    return inventory


def enrich_inventory_include_all(inventory: dict, _: KubernetesCluster) -> dict:
    for _type in ['upgrade', 'remove']:
        packages: dict = inventory['services']['packages'].get(_type)
        if packages is not None:
            packages.setdefault('include', ['*'])

    return inventory


def upgrade_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    # Despite we enrich OS specific section inside enrich_upgrade_inventory(),
    # we still merge global associations section because it has priority during enrichment.
    upgrade_inventory_associations(cluster, inventory, enrich_global=True)
    upgrade_inventory_packages(cluster, inventory)

    return inventory


def migrate_cri_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    if cluster.context.get("initial_procedure") != "migrate_cri":
        return inventory

    procedure_associations = cluster.procedure_inventory.get("packages", {}).get("associations", {})
    # Despite we enrich OS specific section inside enrich_migrate_cri_inventory(),
    # we still merge global associations section because it has priority during enrichment.
    return _enrich_inventory_procedure_associations(cluster, inventory, procedure_associations,
                                                    enrich_global=True)


def cache_package_versions(cluster: KubernetesCluster, inventory: dict, by_initial_nodes: bool = False) -> dict:
    os_ids = cluster.get_os_identifiers()
    different_os = list(set(os_ids.values()))
    if len(different_os) > 1:
        cluster.log.debug(f"Final nodes have different OS families or versions, packages will not be cached. "
                          f"List of (OS family, version): {different_os}")
        return inventory

    os_family = different_os[0][0]
    if os_family in ('unknown', 'unsupported'):
        # For add_node/install procedures we check that OS is supported in prepare.check.system task.
        # For check_iaas procedure it is allowed to have unsupported OS, so skip caching.
        cluster.log.debug("Skip caching of packages for unsupported OS.")
        return inventory

    group = cluster.nodes['all'].get_final_nodes()
    if group.nodes_amount() != group.get_sudo_nodes().nodes_amount():
        # For add_node/install procedures we check that all nodes are sudoers in prepare.check.sudoer task.
        # For check_iaas procedure the nodes might still be not sudoers.
        # Skip caching if any not-sudoer node found.
        cluster.log.debug(f"Some nodes are not sudoers, packages will not be cached.")
        return inventory

    if by_initial_nodes:
        group = group.get_initial_nodes()

    hosts_to_packages = get_all_managed_packages_for_group(group, inventory, by_initial_nodes)
    detected_packages = detect_installed_packages_version_hosts(cluster, hosts_to_packages)

    _cache_package_associations(group, inventory, detected_packages, by_initial_nodes)
    _cache_custom_packages(cluster, inventory, detected_packages, by_initial_nodes)

    cluster.log.debug('Package versions detection finished')
    return inventory


def get_all_managed_packages_for_group(group: NodeGroup, inventory: dict, ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with list of all managed packages for them.
    For associations, only subset of hosts is considered on which the associations are managed by KubeMarine.

    :param group: Group of nodes to get the manager packages for.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account for associations.
                                     Additionally, if true, will skip custom packages.
    :return: List of packages for each relevant host.
    """
    packages_section = inventory['services']['packages']
    hosts_to_packages: Dict[str, List[str]] = {}
    for node in group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        node_associations = packages_section['associations'].get(os_family, {})
        for association_name in node_associations.keys():
            host_packages = get_association_hosts_to_packages(
                node, inventory, association_name, ensured_association_only)

            packages: List[str] = next(iter(host_packages.values()), [])
            hosts_to_packages.setdefault(node.get_host(), []).extend(packages)

    custom_install_packages = inventory['services']['packages'].get('install', {}).get('include', [])
    if not ensured_association_only and custom_install_packages:
        for host in group.get_hosts():
            hosts_to_packages.setdefault(host, []).extend(custom_install_packages)

    return hosts_to_packages


def get_association_hosts_to_packages(group: AbstractGroup[RunResult], inventory: dict, association_name: str,
                                      ensured_association_only: bool = False) \
        -> Dict[str, List[str]]:
    """
    Returns hosts with associated packages list for the specified association name.
    Only subset of hosts is returned on which the association is managed by KubeMarine.

    :param group: Group of nodes to check the applicability of the association.
    :param inventory: Inventory of the cluster. May be different from the inventory of the cluster instance,
                      if used during finalization.
    :param association_name: target association name
    :param ensured_association_only: Specify whether to take 'cache_versions' property into account.
    :return: List of packages for each relevant host.
    """
    cluster: KubernetesCluster = group.cluster

    packages_section = inventory['services']['packages']
    if not packages_section['mandatory'].get(association_name, True):
        return {}

    hosts_to_packages = {}

    if association_name == 'unzip':
        from kubemarine import thirdparties
        relevant_group: AbstractGroup[RunResult] = thirdparties.get_group_require_unzip(cluster, inventory)
    else:
        groups = cluster.globals['packages']['common_associations'].get(association_name, {}).get('groups', [])
        relevant_group = cluster.make_group_from_roles(groups)

    if association_name in ('docker', 'containerd') \
            and association_name != inventory['services']['cri']['containerRuntime']:
        relevant_group = cluster.make_group([])

    relevant_group = relevant_group.intersection_group(group)

    global_cache_versions = packages_section['cache_versions']
    for node in relevant_group.get_ordered_members_list():
        os_family = node.get_nodes_os()
        package_associations = packages_section['associations'].get(os_family, {}).get(association_name, {})
        packages = package_associations.get('package_name', [])

        if isinstance(packages, str):
            packages = [packages]

        if ensured_association_only and not (global_cache_versions and package_associations.get('cache_versions', True)):
            packages = []

        if packages:
            hosts_to_packages[node.get_host()] = packages

    return hosts_to_packages


def _cache_package_associations(group: NodeGroup, inventory: dict,
                                detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool) -> None:
    cluster: KubernetesCluster = group.cluster
    associations = inventory['services']['packages']['associations'][cluster.get_os_family()]
    for association_name, associated_params in associations.items():
        hosts_to_packages = get_association_hosts_to_packages(
            group, inventory, association_name, ensured_association_only)
        if not hosts_to_packages:
            continue

        # Since all nodes have the same OS family in this case,
        # the packages list is the same for all relevant hosts, so take any available.
        packages_list = next(iter(hosts_to_packages.values()))

        final_packages_list = []
        for package in packages_list:
            final_package = _detect_final_package(cluster, detected_packages, package, ensured_association_only)
            final_packages_list.append(final_package)

        # if non-multiple value, then convert to simple string
        # packages can contain multiple package values, like docker package
        # (it has docker-ce, docker-cli and containerd.io packages for installation)
        if len(final_packages_list) == 1:
            associated_params['package_name'] = final_packages_list[0]
        else:
            associated_params['package_name'] = final_packages_list


def _cache_custom_packages(cluster: KubernetesCluster, inventory: dict,
                           detected_packages: Dict[str, Dict[str, List]], ensured_association_only: bool) -> None:
    if ensured_association_only:
        return
    # packages from direct installation section
    custom_install_packages = inventory['services']['packages'].get('install', {})
    if custom_install_packages.get('include', []):
        final_packages_list = []
        for package in custom_install_packages['include']:
            final_package = _detect_final_package(cluster, detected_packages, package, False)
            final_packages_list.append(final_package)
        custom_install_packages['include'] = final_packages_list


def _detect_final_package(cluster: KubernetesCluster, detected_packages: Dict[str, Dict[str, List]],
                          package: str, ensured_association_only: bool) -> str:
    # add package version to list only if it was found as installed
    detected_package_versions = list(filter(lambda version: "not installed" not in version,
                                            detected_packages[package].keys()))

    # if there no versions detected, then return default package from inventory
    if not detected_package_versions:
        return package
    elif len(detected_package_versions) > 1:
        if ensured_association_only:
            raise Exception(ERROR_MULTIPLE_PACKAGE_VERSIONS_DETECTED % (str(detected_packages[package]), package))
        else:
            cluster.log.warning(
                f"Multiple package versions detected {detected_packages[package]} for package '{package}'. "
                f"Use default package '{package}' from inventory.")
            # return default package from inventory if multiple versions detected
            return package
    else:
        return detected_package_versions[0]


def remove_unused_os_family_associations(cluster: KubernetesCluster, inventory: dict) -> dict:
    final_nodes = cluster.nodes['all'].get_final_nodes()
    for os_family in get_associations_os_family_keys():
        # Do not remove OS family associations section in finalized inventory if any node has this OS family.
        if final_nodes.get_subgroup_with_os(os_family).is_empty():
            del inventory['services']['packages']['associations'][os_family]

    return inventory


def get_associations_os_family_keys() -> Set[str]:
    return {'debian', 'rhel', 'rhel8', 'rhel9'}


def get_compatibility_version_key(os_family: str) -> str:
    """
    Get os-specific version key to be used in software compatibility map.
    :param os_family: one of supported OS families
    :return: String to use as version key.
    """
    if os_family in get_associations_os_family_keys():
        return f"version_{os_family}"
    else:
        raise ValueError(f"Unsupported {os_family!r} OS family")


class PackageManager(Protocol):
    def ls_repofiles(self, group: NodeGroup) -> RunnersGroupResult: ...

    def backup_repo(self, group: NodeGroup) -> Optional[RunnersGroupResult]: ...

    def add_repo(self, group: NodeGroup, repo_data: Union[List[str], Dict[str, dict], str]) -> RunnersGroupResult: ...

    def get_repo_file_name(self) -> str: ...

    def create_repo_file(self, group: AbstractGroup[RunResult],
                         repo_data: Union[List[str], Dict[str, dict], str],
                         repo_file: str) -> None: ...

    def clean(self, group: NodeGroup) -> RunnersGroupResult: ...

    def get_install_cmd(self, include: Union[str, List[str]], exclude: Union[str, List[str]] = None) -> str: ...

    def install(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
                exclude: Union[str, List[str]] = None,
                callback: Callback = None) -> GROUP_RUN_TYPE: ...

    def remove(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
               exclude: Union[str, List[str]] = None,
               warn: bool = False, hide: bool = True) -> GROUP_RUN_TYPE: ...

    def upgrade(self, group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
                exclude: Union[str, List[str]] = None) -> GROUP_RUN_TYPE: ...

    def no_changes_found(self, action: str, result: RunnersResult) -> bool: ...

    def search(self, group: DeferredGroup, package: str, callback: Callback = None) -> Token: ...


def get_package_manager(group: AbstractGroup[GROUP_RUN_TYPE]) -> PackageManager:
    os_family = group.get_nodes_os()

    if os_family in ['rhel', 'rhel8', 'rhel9']:
        return yum
    elif os_family == 'debian':
        return apt

    raise Exception('Failed to return package manager for unknown or multiple OS')


def ls_repofiles(group: NodeGroup) -> RunnersGroupResult:
    return get_package_manager(group).ls_repofiles(group)


def backup_repo(group: NodeGroup) -> Optional[RunnersGroupResult]:
    return get_package_manager(group).backup_repo(group)


def add_repo(group: NodeGroup, repo_data: Union[List[str], dict, str]) -> RunnersGroupResult:
    return get_package_manager(group).add_repo(group, repo_data)


def get_repo_filename(group: AbstractGroup[RunResult]) -> str:
    return get_package_manager(group).get_repo_file_name()


def create_repo_file(group: AbstractGroup[RunResult], repo_data: Union[List[str], dict, str], repo_file: str) -> None:
    get_package_manager(group).create_repo_file(group, repo_data, repo_file)


def clean(group: NodeGroup) -> RunnersGroupResult:
    return get_package_manager(group).clean(group)


def install(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None,
            callback: Callback = None) -> GROUP_RUN_TYPE:
    return get_package_manager(group).install(group, include, exclude, callback)


def remove(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None, exclude: Union[str, List[str]] = None,
           warn: bool = False, hide: bool = True) -> GROUP_RUN_TYPE:
    return get_package_manager(group).remove(group, include, exclude, warn=warn, hide=hide)


def upgrade(group: AbstractGroup[GROUP_RUN_TYPE], include: Union[str, List[str]] = None,
            exclude: Union[str, List[str]] = None) -> GROUP_RUN_TYPE:
    return get_package_manager(group).upgrade(group, include, exclude)


def no_changes_found(group: NodeGroup, action: str, result: RunnersResult) -> bool:
    pkg_mgr = get_package_manager(group)
    return pkg_mgr.no_changes_found(action, result)


def search_package(group: DeferredGroup, package: str, callback: Callback = None) -> Token:
    return get_package_manager(group).search(group, package, callback)


def get_detect_package_version_cmd(os_family: str, package_name: str) -> str:
    if os_family in ["rhel", "rhel8", "rhel9"]:
        cmd = r"rpm -q %s" % package_name
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W %s" % package_name

    return cmd


def _detect_installed_package_version(group: DeferredGroup, package: str, collector: CollectorCallback) -> Token:
    """
    Detect package versions for each host on remote group
    :param group: Group of nodes, where package should be found
    :param package: package name, which version should be detected (eg. 'containerd')
    :return: NodeGroupResults with package version on each host

    Method generates different package query for different OS.

    Note: for Ubuntu/Debian some packages returns multiline results for some queries
    (for example docker-ce* returns docker-ce and docker-ce-cli).
    """

    os_family = group.get_nodes_os()
    package_name = get_package_name(os_family, package)

    cmd = get_detect_package_version_cmd(os_family, package_name)
    return group.sudo(cmd, warn=True, callback=collector)


def _parse_node_detected_package(result: RunnersResult, package: str) -> str:
    node_detected_package = result.stdout.strip() + result.stderr.strip()
    # consider version, which ended with special symbol = or - as not installed
    # (it is possible in some cases to receive "containerd=" version)
    if "not installed" in node_detected_package or "no packages found" in node_detected_package \
            or node_detected_package[-1] == '=' or node_detected_package[-1] == '-':
        node_detected_package = f"not installed {package}"

    return node_detected_package


def detect_installed_packages_version_hosts(
        cluster: KubernetesCluster, hosts_to_packages: Mapping[str, Union[str, List[str]]]
) -> Dict[str, Dict[str, List[str]]]:
    """
    Detect grouped packages versions for specified list of packages for each remote host.

    :param cluster: KubernetesCluster instance
    :param hosts_to_packages: Remote hosts with list of packages to detect versions.
    :return: Dictionary with grouped versions for each queried package, pointing to list of hosts,
        e.g. {"foo" -> {"foo-1": [host1, host2]}, "bar" -> {"bar-1": [host1], "bar-2": [host2]}}
    """
    hosts_to_packages_dedup = {}
    for host, packages_list in hosts_to_packages.items():
        if isinstance(packages_list, str):
            packages_list = [packages_list]
        # deduplicate
        hosts_to_packages_dedup[host] = list(set(packages_list))

    collector = CollectorCallback(cluster)
    with cluster.make_group(hosts_to_packages).new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            for package in hosts_to_packages_dedup[node.get_host()]:
                _detect_installed_package_version(node, package, collector)

    results: Dict[str, Dict[str, List]] = {}

    for host, multiple_results in collector.results.items():
        for i, result in enumerate(multiple_results):
            package = hosts_to_packages_dedup[host][i]
            node_detected_package = _parse_node_detected_package(result, package)
            results.setdefault(package, {}).setdefault(node_detected_package, []).append(host)

    return results


def get_package_version_separator(os_family: str) -> str:
    return '=' if os_family == 'debian' else '-'


def get_package_name(os_family: str, package: str) -> str:
    """
    Return the pure package name, without any part of version
    """

    import re

    package_name = ""

    if package:
        if os_family in ["rhel", "rhel8", "rhel9"]:
            # regexp is needed to split package and its version, the pattern start with '-' then should be number or '*'
            package_name = re.split(r'-[\d,\*]', package)[0]
        else:
            # in ubuntu it is much easier to parse package name
            package_name = package.split("=")[0]

    return package_name



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from abc import ABC, abstractmethod
from functools import total_ordering

from kubemarine.core.action import Action


@total_ordering
class _Priority(enum.Enum):
    INVENTORY_ONLY = 0
    "The patch should only change the inventory. Enrichment is prohibited."

    SOFTWARE_UPGRADE = 1
    "This is a service patch that should be instantiated only automatically by migrate_kubemarine.py"

    REGULAR = 2
    """
    The patch can access and make some operations on the cluster.
    Changes in the inventory are possible, but they should not affect the software upgrade procedure.
    """

    def __lt__(self, other: '_Priority') -> bool:
        return self.value < other.value


class Patch(ABC):
    def __init__(self, identifier: str):
        self.identifier = identifier

    @abstractmethod
    def priority(self) -> _Priority:
        pass

    @property
    @abstractmethod
    def action(self) -> Action:
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        pass


class InventoryOnlyPatch(Patch, ABC):
    """
    The patch should only change the inventory.
    Enrichment is prohibited. Calling DynamicResources.cluster() is prohibited.
    Patches if this type are executed first.
    """

    def priority(self) -> _Priority:
        return _Priority.INVENTORY_ONLY


class _SoftwareUpgradePatch(Patch, ABC):
    """This is a service patch that should be extended only by predefined set of classes inside migrate_kubemarine.py"""

    def priority(self) -> _Priority:
        return _Priority.SOFTWARE_UPGRADE


class RegularPatch(Patch, ABC):
    """
    The patch can access and make some operations on the cluster.
    Changes in the inventory are possible, but they should not affect the software upgrade procedure.
    Patches if this type are executed last.
    """

    def priority(self) -> _Priority:
        return _Priority.REGULAR



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict

import io
import uuid
from typing import List

from kubemarine import kubernetes, etcd, thirdparties, cri
from kubemarine.core import flow
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.core.resources import DynamicResources
from kubemarine.cri import docker
from kubemarine.procedures import install
from kubemarine import packages


def migrate_cri(cluster: KubernetesCluster) -> None:
    _migrate_cri(cluster, cluster.nodes["worker"].exclude_group(cluster.nodes["control-plane"])
                 .get_ordered_members_list())
    _migrate_cri(cluster, cluster.nodes["control-plane"].get_ordered_members_list())


def _migrate_cri(cluster: KubernetesCluster, node_group: List[NodeGroup]) -> None:
    """
    Migrate CRI from docker to already installed containerd.
    This method works node-by-node, configuring kubelet to use containerd.
    :param cluster: main object describing a cluster
    :param node_group: group of nodes to migrate
    """

    for node in node_group:
        node_config = node.get_config()
        node_name = node.get_node_name()
        is_control_plane = "control-plane" in node_config["roles"]
        if is_control_plane:
            control_plane = node
        else:
            control_plane = cluster.nodes["control-plane"].get_first_member()

        cluster.log.debug(f'Updating thirdparties for node "{node_name}"...')
        thirdparties.install_all_thirparties(node)

        version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
        cluster.log.debug("Migrating \"%s\"..." % node_name)
        drain_cmd = kubernetes.prepare_drain_command(cluster, node_name, disable_eviction=True)
        control_plane.sudo(drain_cmd, hide=False)

        kubeadm_flags_file = "/var/lib/kubelet/kubeadm-flags.env"
        kubeadm_flags = node.sudo(f"cat {kubeadm_flags_file}").get_simple_out()

        #Removing the --network-plugin=cni switch after the cri migration procedure that was used to run Docker on the cluster.
        #Support for this key has been removed in kubernetes 1.24.
        if kubeadm_flags.find('--network-plugin=cni') != -1:
            kubeadm_flags = kubeadm_flags.replace('--network-plugin=cni', '')

        kubeadm_flags = edit_config(kubeadm_flags)

        node.put(io.StringIO(kubeadm_flags), kubeadm_flags_file, backup=True, sudo=True)

        node.sudo("systemctl stop kubelet")
        docker.prune(node)

        docker_associations = cluster.get_associations_for_node(node.get_host(), 'docker')
        node.sudo(f"systemctl disable {docker_associations['service_name']} --now; "
                  "sudo sh -c 'rm -rf /var/lib/docker/*'")

        cluster.log.debug('Reinstalling CRI...')
        cri.install(node)
        cri.configure(node)

        cluster.log.debug(f'CRI configured! Restoring pods on node "{node_name}"')

        # if there is a disk for docker in "/etc/fstab", then use this disk for containerd
        docker_disk_result = node.sudo("cat /etc/fstab | grep ' /var/lib/docker '", warn=True)
        docker_disk = list(docker_disk_result.values())[0].stdout.strip()
        if docker_disk:
            node.sudo(
                "umount /var/lib/docker && "
                "sudo sed -i 's/ \/var\/lib\/docker / \/var\/lib\/containerd /' /etc/fstab && "
                "sudo sh -c 'rm -rf /var/lib/containerd/*' && "
                "sudo mount -a && "
                "sudo systemctl restart containerd")

        # flushing iptables to delete old cri's rules,
        # existence of those rules could lead to services unreachable
        node.sudo(
            "sudo iptables -t nat -F && "
            "sudo iptables -t raw -F && "
            "sudo iptables -t filter -F && "
            # hotfix for Ubuntu 22.04
            "sudo systemctl stop kubepods-burstable.slice || true && "
            "sudo systemctl restart containerd && "
            # start kubelet
            "sudo systemctl restart kubelet")

        if is_control_plane:
            kubernetes.wait_uncordon(node)
        else:
            control_plane.sudo(f"kubectl uncordon {node_name}", hide=False)

        if is_control_plane:
            kubernetes.wait_for_any_pods(cluster, node, apply_filter=node_name)
            # check ETCD health
            etcd.wait_for_health(cluster, node)

        packages_list = []
        for package_name in docker_associations['package_name']:
            if not package_name.startswith('containerd'):
                packages_list.append(package_name)
        cluster.log.warning("The following packages will be removed: %s" % packages_list)
        if packages_list:
            packages.remove(node, include=packages_list, warn=True, hide=False)

        # change annotation for cri-socket
        control_plane.sudo(f"sudo kubectl annotate node {node_name} "
                           f"--overwrite kubeadm.alpha.kubernetes.io/cri-socket=/run/containerd/containerd.sock")

        # delete docker socket
        node.sudo("rm -rf /var/run/docker.sock", hide=False)


def release_calico_leaked_ips(cluster: KubernetesCluster) -> None:
    """
    During drain command we ignore daemon sets, as result this such pods as ingress-nginx-controller arent't deleted before migration.
    For this reason their ips can stay in calico ipam despite they aren't used. You can check this, if you run "calicoctl ipam check --show-problem-ips" right after apply_new_cri task.
    Those ips are cleaned by calico garbage collector, but it can take about 20 minutes.
    This task releases problem ips with force.
    """
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    cluster.log.debug("Getting leaked ips...")
    random_report_name = "/tmp/%s.json" % uuid.uuid4().hex
    result = first_control_plane.sudo(f"calicoctl ipam check --show-problem-ips -o {random_report_name} | grep 'leaked' || true", hide=False)
    leaked_ips = result.get_simple_out()
    leaked_ips_count = leaked_ips.count('leaked')
    cluster.log.debug(f"Found {leaked_ips_count} leaked ips")
    if leaked_ips_count != 0:
        first_control_plane.sudo(f"calicoctl ipam release --from-report={random_report_name} --force", hide=False)
        cluster.log.debug("Leaked ips was released")
    first_control_plane.sudo(f"rm {random_report_name}", hide=False)


def edit_config(kubeadm_flags: str) -> str:
    kubeadm_flags = kubernetes.config_changer(kubeadm_flags, "--container-runtime=remote")
    return kubernetes.config_changer(kubeadm_flags,
                                     "--container-runtime-endpoint=unix:///run/containerd/containerd.sock")


tasks = OrderedDict({
    "add_repos": install.system_prepare_package_manager_configure,
    "apply_new_cri": migrate_cri,
    "release_calico_ipam_leacked_ips": release_calico_leaked_ips
})


class MigrateCRIAction(Action):
    def __init__(self) -> None:
        super().__init__('migrate cri', recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks)
        res.make_final_inventory()


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
        Script for automated migration from docker to containerd.

        How to use:

        '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure="migrate_cri")
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([MigrateCRIAction()]).run_flow(context)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import List

from kubemarine import kubernetes, haproxy, keepalived
from kubemarine.core import flow, summary
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.core.resources import DynamicResources
from kubemarine.procedures import install, add_node


def get_active_nodes(node_type: str, cluster: KubernetesCluster) -> NodeGroup:
    all_nodes = cluster.make_group_from_roles([node_type]).get_nodes_for_removal()
    if all_nodes.is_empty():
        cluster.log.debug("Skipped - no %s to remove" % node_type)
        return all_nodes
    active_nodes = all_nodes.get_online_nodes(True)
    disabled_nodes = all_nodes.exclude_group(active_nodes)
    if active_nodes.is_empty():
        cluster.log.debug("Skipped - %s nodes are inactive: %s" % (node_type, ", ".join(disabled_nodes.nodes)))
        return active_nodes
    if not disabled_nodes.is_empty():
        cluster.log.debug("Partly Skipped - several %s nodes are inactive: %s"
                          % (node_type, ", ".join(disabled_nodes.nodes)))
    return active_nodes


def loadbalancer_remove_haproxy(cluster: KubernetesCluster) -> None:
    nodes = get_active_nodes("balancer", cluster)
    if nodes.is_empty():
        return
    nodes.call(haproxy.disable)


def loadbalancer_remove_keepalived(cluster: KubernetesCluster) -> None:
    nodes = get_active_nodes("keepalived", cluster)
    if nodes.is_empty():
        return
    nodes.call(keepalived.disable)


def remove_kubernetes_nodes(cluster: KubernetesCluster) -> None:
    group = cluster.make_group_from_roles(['control-plane', 'worker']).get_nodes_for_removal()

    if group.is_empty():
        cluster.log.debug("No kubernetes nodes to perform")
        return

    group.call(kubernetes.reset_installation_env)
    kubernetes.schedule_running_nodes_report(cluster)


def remove_node_finalize_inventory(cluster: KubernetesCluster, inventory_to_finalize: dict) -> dict:
    if cluster.context.get('initial_procedure') != 'remove_node':
        return inventory_to_finalize

    final_nodes = cluster.nodes['all'].get_final_nodes()

    is_finalization = any('remove_node' in node['roles'] for node in inventory_to_finalize['nodes'])

    if not is_finalization:
        kubernetes.remove_node_enrichment(inventory_to_finalize, cluster)

    # Do not remove VRRP IPs and do not change their assigned hosts.
    # If the assigned host does not exist or is not a balancer, it will be just skipped.
    # Though it is necessary to remove hosts from the VRRP IP of finalized inventory if they were enriched ourselves.
    if is_finalization:
        for i, item in enumerate(inventory_to_finalize.get('vrrp_ips', [])):
            raw_item = cluster.raw_inventory['vrrp_ips'][i]
            # If redefined, it was not enriched. See keepalived.enrich_inventory_apply_defaults().
            if not isinstance(raw_item, str) and 'hosts' in raw_item:
                continue

            item['hosts'] = [host for host in item['hosts'] if final_nodes.has_node(host['name'])]

    # remove nodes from inventory if they in nodes for removal
    size = len(inventory_to_finalize['nodes'])
    for i in range(size):
        for j, node in enumerate(inventory_to_finalize['nodes']):
            if 'remove_node' in node['roles']:
                del inventory_to_finalize['nodes'][j]
                break

    return inventory_to_finalize


tasks = OrderedDict({
    "loadbalancer": {
        "remove": {
            "haproxy": loadbalancer_remove_haproxy,
            "keepalived": loadbalancer_remove_keepalived
        },
        "configure": {
            "haproxy": install.deploy_loadbalancer_haproxy_configure,
            "keepalived": install.deploy_loadbalancer_keepalived_configure
        },
    },
    "update": {
        "etc_hosts": install.system_prepare_dns_etc_hosts,
        "coredns": install.deploy_coredns,
        "plugins": add_node.redeploy_plugins_if_needed,
    },
    "remove_kubernetes_nodes": remove_kubernetes_nodes,
    "overview": install.overview,
})


cumulative_points = {
    summary.exec_delayed: [
        flow.END_OF_TASKS
    ]
}


class RemoveNodeAction(Action):
    def __init__(self) -> None:
        super().__init__('remove node', recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        flow.run_tasks(res, tasks, cumulative_points=cumulative_points)
        res.make_final_inventory()


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for removing node from Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='remove_node')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([RemoveNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import math
import os
import time
import uuid
from contextlib import contextmanager
from copy import deepcopy
from typing import List, Dict, Tuple, Iterator, Any, Optional

import yaml
from jinja2 import Template
import ipaddress

from kubemarine import system, plugins, admission, etcd, packages
from kubemarine.core import utils, static, summary, log, errors
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.executor import Token
from kubemarine.core.group import (
    NodeGroup, AbstractGroup, DeferredGroup,
    NodeConfig, RunnersGroupResult, RunResult, CollectorCallback
)
from kubemarine.core.errors import KME
from kubemarine.cri import containerd

ERROR_DOWNGRADE='Kubernetes old version \"%s\" is greater than new one \"%s\"'
ERROR_SAME='Kubernetes old version \"%s\" is the same as new one \"%s\"'
ERROR_MAJOR_RANGE_EXCEEDED='Major version \"%s\" rises to new \"%s\" more than one'
ERROR_MINOR_RANGE_EXCEEDED='Minor version \"%s\" rises to new \"%s\" more than one'


def add_node_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get('initial_procedure') != 'add_node':
        return inventory

    # adding role "new_node" for all specified new nodes and putting these nodes to all "nodes" list
    for new_node in cluster.procedure_inventory.get("nodes", []):
        # deepcopy is necessary, otherwise role append will happen in procedure_inventory too
        node = deepcopy(new_node)
        node["roles"].append("add_node")
        inventory["nodes"].append(node)

    # If "vrrp_ips" section is ever supported when adding node,
    # It will be necessary to more accurately install and reconfigure the keepalived on existing nodes.

    # if "vrrp_ips" in cluster.procedure_inventory:
    #     utils.merge_vrrp_ips(cluster.procedure_inventory, inventory)

    return inventory


def remove_node_enrichment(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get('initial_procedure') != 'remove_node':
        return inventory

    # adding role "remove_node" for all specified nodes
    node_names_to_remove = [node['name'] for node in cluster.procedure_inventory.get("nodes", [])]
    for node_remove in node_names_to_remove:
        for i, node in enumerate(inventory['nodes']):
            # Inventory is not compiled at this step.
            # Expecting that the names are not jinja, or the same jinja expressions.
            if node['name'] == node_remove:
                node['roles'].append('remove_node')
                break
        else:
            raise Exception(f"Failed to find node to remove {node_remove} among existing nodes")

    return inventory


def enrich_upgrade_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get('initial_procedure') == 'upgrade':
        cluster.context['initial_kubernetes_version'] = inventory['services']['kubeadm']['kubernetesVersion']

        cluster.log.info(
            '------------------------------------------\nUPGRADING KUBERNETES %s ⭢ %s\n------------------------------------------' % (
            cluster.context['initial_kubernetes_version'], cluster.context['upgrade_version']))

    return generic_upgrade_inventory(cluster, inventory)


def upgrade_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    return generic_upgrade_inventory(cluster, inventory)


def generic_upgrade_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    if cluster.context.get("initial_procedure") != "upgrade":
        return inventory

    upgrade_version = cluster.context.get("upgrade_version")
    inventory.setdefault("services", {}).setdefault("kubeadm", {})['kubernetesVersion'] = upgrade_version
    return inventory


def enrich_restore_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:
    if cluster.context.get("initial_procedure") != "restore":
        return inventory

    logger = cluster.log
    kubernetes_descriptor = cluster.context['backup_descriptor'].setdefault('kubernetes', {})
    initial_kubernetes_version = get_initial_kubernetes_version(inventory)
    backup_kubernetes_version = kubernetes_descriptor.get('version')
    if not backup_kubernetes_version:
        logger.warning("Not possible to verify Kubernetes version, as descriptor does not contain 'kubernetes.version'")
        backup_kubernetes_version = initial_kubernetes_version

    if backup_kubernetes_version != initial_kubernetes_version:
        logger.warning('Installed kubernetes version does not match version from backup')
        verify_allowed_version(backup_kubernetes_version)

    kubernetes_descriptor['version'] = backup_kubernetes_version
    return restore_finalize_inventory(cluster, inventory)


def restore_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    if cluster.context.get("initial_procedure") != "restore":
        return inventory

    target_kubernetes_version = cluster.context['backup_descriptor']['kubernetes']['version']
    inventory.setdefault("services", {}).setdefault("kubeadm", {})['kubernetesVersion'] = target_kubernetes_version
    return inventory


def enrich_inventory(inventory: dict, _: KubernetesCluster) -> dict:
    kubeadm = inventory['services']['kubeadm']
    kubeadm['dns'].setdefault('imageRepository', f"{kubeadm['imageRepository']}/coredns")

    enriched_certsans = []

    for node in inventory["nodes"]:
        if ('balancer' in node['roles'] or 'control-plane' in node['roles']) and 'remove_node' not in node['roles']:
            enriched_certsans.extend([node['name'], node['internal_address']])
            if node.get('address') is not None:
                enriched_certsans.append(node['address'])

    # The VRRP IP may be actually unused, but let's add it because it is probably specified to be used in the future.
    for item in inventory["vrrp_ips"]:
        enriched_certsans.append(item['ip'])
        if item.get("floating_ip"):
            enriched_certsans.append(item["floating_ip"])

    if inventory.get("public_cluster_ip"):
        enriched_certsans.append(inventory["public_cluster_ip"])

    certsans = inventory["services"]["kubeadm"]['apiServer']['certSANs']

    # do not overwrite apiServer.certSANs, but append - may be user specified something already there?
    for name in enriched_certsans:
        if name not in certsans:
            certsans.append(name)

    any_worker_found = False

    # validating node labels and configuring additional labels
    for node in inventory["nodes"]:
        if "control-plane" not in node["roles"] and "worker" not in node["roles"]:
            if "labels" in node:
                raise Exception("Only 'worker' or 'control-plane' nodes can have labels, "
                                "but found label on %s, roles: %s" % (node["name"], node["roles"]))
            if "taints" in node:
                raise Exception("Only 'worker' or 'control-plane' nodes can have taints, "
                                "but found taints on %s, roles: %s" % (node["name"], node["roles"]))
            continue

        if "worker" in node["roles"]:
            any_worker_found = True

            if "labels" not in node:
                node["labels"] = {}
            node["labels"]["node-role.kubernetes.io/worker"] = "worker"
            
    # Validate the provided podSubnet IP address
    pod_subnet = inventory.get('services', {}).get('kubeadm', {}).get('networking', {}).get('podSubnet')
    try:
        ip_network = ipaddress.ip_network(pod_subnet)
        if ip_network.version not in [4, 6]:
            raise ValueError(f"Invalid podSubnet IP address: {pod_subnet}")
    except ValueError:
        raise ValueError(f"Invalid podSubnet IP address: {pod_subnet}")

    # Validate the provided serviceSubnet IP address
    service_subnet = inventory.get('services', {}).get('kubeadm', {}).get('networking', {}).get('serviceSubnet')
    try:
        ip_network = ipaddress.ip_network(service_subnet)
        if ip_network.version not in [4, 6]:
            raise ValueError(f"Invalid serviceSubnet IP address: {service_subnet}")
    except ValueError:
        raise ValueError(f"Invalid serviceSubnet IP address: {service_subnet}")

    # validate nodes in kubeadm_patches (groups are validated with JSON schema)
    for node in inventory["nodes"]:
        for control_plane_item in inventory["services"]["kubeadm_patches"]:
            for i in inventory["services"]["kubeadm_patches"][control_plane_item]:
                if i.get('nodes') is not None:
                    for n in i['nodes']:
                        if node['name'] == n:
                            if control_plane_item == 'kubelet' and 'control-plane' not in node['roles'] and 'worker' not in node['roles']:
                                raise Exception("%s patch can be uploaded only to control-plane or worker nodes" % control_plane_item)
                            if control_plane_item != 'kubelet' and ('control-plane' not in node['roles']):
                                raise Exception("%s patch can be uploaded only to control-plane nodes" % control_plane_item)

    if not any_worker_found:
        raise KME("KME0004")

    # check ignorePreflightErrors value and add mandatory errors from defaults.yaml if they're absent
    default_preflight_errors = static.DEFAULTS["services"]["kubeadm_flags"]["ignorePreflightErrors"].split(",")
    preflight_errors = inventory["services"]["kubeadm_flags"]["ignorePreflightErrors"].split(",")

    preflight_errors.extend(default_preflight_errors)
    inventory["services"]["kubeadm_flags"]["ignorePreflightErrors"] = ",".join(set(preflight_errors))

    return inventory


def reset_installation_env(group: NodeGroup) -> Optional[RunnersGroupResult]:
    log = group.cluster.log

    log.debug("Cleaning up previous installation...")

    cluster: KubernetesCluster = group.cluster

    drain_timeout = cluster.procedure_inventory.get('drain_timeout')
    grace_period = cluster.procedure_inventory.get('grace_period')

    # if we perform "add" or "remove" node procedure
    # then we need to additionally perform "drain" and "delete" during reset
    nodes_for_draining = cluster.make_group([])

    # perform FULL reset only for "add" or "remove" procedures
    # do not perform full reset on cluster (re)installation, it could hang on last etcd member
    # nodes should be deleted only during "add" or "remove" procedures
    is_add_or_remove_procedure = True

    nodes_for_manual_etcd_remove = cluster.make_group([])

    if not group.get_nodes_for_removal().is_empty():
        # this is remove_node procedure
        active_nodes = group.get_online_nodes(True)

        # We need to manually remove members from etcd for "remove" procedure,
        # only if corresponding nodes are not active.
        # Otherwise, they will be removed by "kubeadm reset" command.
        nodes_for_manual_etcd_remove = group.exclude_group(active_nodes)

        # kubectl drain command hands on till timeout is exceeded for nodes which are off
        # so we should drain only active nodes
        nodes_for_draining = active_nodes
    else:
        # in other case we consider all nodes are active
        active_nodes = group

        if not group.get_new_nodes().is_empty():
            # this is add_node procedure
            nodes_for_draining = group
        else:
            # this is install procedure
            is_add_or_remove_procedure = False

    if not nodes_for_manual_etcd_remove.is_empty():
        log.warning(f"Nodes {nodes_for_manual_etcd_remove.get_hosts()} are considered as not active. "
                    "Full cleanup procedure cannot be performed. "
                    "Corresponding members will be removed from etcd manually.")
        etcd.remove_members(nodes_for_manual_etcd_remove)

    if not nodes_for_draining.is_empty():
        drain_nodes(nodes_for_draining, drain_timeout=drain_timeout, grace_period=grace_period)

    if is_add_or_remove_procedure and not active_nodes.is_empty():
        log.verbose(f"Resetting kubeadm on nodes {active_nodes.get_hosts()} ...")
        result = active_nodes.sudo('sudo kubeadm reset -f')
        log.debug("Kubeadm successfully reset:\n%s" % result)

    if not active_nodes.is_empty():
        log.verbose(f"Cleaning nodes {active_nodes.get_hosts()} ...")
        # bash semicolon mark will avoid script from exiting and will resume the execution
        result = active_nodes.sudo(
            'sudo kubeadm reset phase cleanup-node; '  # it is required to "cleanup-node" for all procedures
            'sudo systemctl stop kubelet; '
            'sudo rm -rf /etc/kubernetes/manifests /var/lib/kubelet/pki /var/lib/etcd /etc/kubernetes/patches; '
            'sudo mkdir -p /etc/kubernetes/manifests; ', warn=True)

        # Disabled initial prune for images prepull feature. Need analysis for possible negative impact.
        # result.update(cri.prune(active_nodes, all_implementations=True))

        log.debug(f"Nodes {active_nodes.get_hosts()} cleaned up successfully:\n" + "%s" % result)

    if is_add_or_remove_procedure:
        return delete_nodes(group)

    return None


def drain_nodes(group: NodeGroup, disable_eviction: bool = False,
                drain_timeout: int = None, grace_period: int = None) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    control_plane = cluster.nodes['control-plane'].get_final_nodes().get_first_member()
    result = control_plane.sudo("kubectl get nodes -o custom-columns=NAME:.metadata.name")

    stdout = list(result.values())[0].stdout
    log.verbose("Detected the following nodes in cluster:\n%s" % stdout)

    for node in group.get_ordered_members_list():
        node_name = node.get_node_name()
        if node_name in stdout:
            log.debug("Draining node %s..." % node_name)
            drain_cmd = prepare_drain_command(
                cluster, node_name,
                disable_eviction=disable_eviction, drain_timeout=drain_timeout, grace_period=grace_period)
            control_plane.sudo(drain_cmd, hide=False)
        else:
            log.warning("Node %s is not found in cluster and can't be drained" % node_name)

    return control_plane.sudo("kubectl get nodes")


def delete_nodes(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    control_plane = cluster.nodes['control-plane'].get_final_nodes().get_first_member()
    result = control_plane.sudo("kubectl get nodes -o custom-columns=NAME:.metadata.name")

    stdout = list(result.values())[0].stdout
    log.verbose("Detected the following nodes in cluster:\n%s" % stdout)

    for node in group.get_ordered_members_list():
        node_name = node.get_node_name()
        if node_name in stdout:
            log.debug("Deleting node %s from the cluster..." % node_name)
            control_plane.sudo("kubectl delete node %s" % node_name, hide=False)
        else:
            log.warning("Node %s is not found in cluster and can't be removed" % node_name)

    return control_plane.sudo("kubectl get nodes")


def install(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    with group.new_executor() as exe:
        log.debug("Making systemd unit...")
        for node in exe.group.get_ordered_members_list():
            node.sudo('rm -rf /etc/systemd/system/kubelet*')
            template = Template(utils.read_internal('templates/kubelet.service.j2')).render(
                hostname=node.get_node_name())
            log.debug("Uploading to '%s'..." % node.get_host())
            node.put(io.StringIO(template + "\n"), '/etc/systemd/system/kubelet.service', sudo=True)
            node.sudo("chmod 600 /etc/systemd/system/kubelet.service")

        log.debug("\nReloading systemd daemon...")
        system.reload_systemctl(exe.group)
        exe.group.sudo('systemctl enable kubelet')

    return group.sudo('systemctl status kubelet', warn=True)


def join_other_control_planes(group: NodeGroup) -> RunnersGroupResult:
    other_control_planes_group = group.get_ordered_members_list()[1:]

    join_dict = group.cluster.context["join_dict"]
    for node in other_control_planes_group:
        join_control_plane(group.cluster, node, join_dict)

    group.cluster.log.debug("Verifying installation...")
    first_control_plane = group.get_first_member()
    return first_control_plane.sudo("kubectl get pods --all-namespaces -o=wide")


def join_new_control_plane(group: NodeGroup) -> None:
    join_dict = get_join_dict(group)
    for node in group.get_ordered_members_list():
        join_control_plane(group.cluster, node, join_dict)


def join_control_plane(cluster: KubernetesCluster, node: NodeGroup, join_dict: dict) -> None:
    log = cluster.log
    node_config = node.get_config()
    node_name = node.get_node_name()
    defer = node.new_defer()

    join_config: dict = {
        'apiVersion': cluster.inventory["services"]["kubeadm"]['apiVersion'],
        'kind': 'JoinConfiguration',
        'discovery': {
            'bootstrapToken': {
                'apiServerEndpoint': cluster.inventory["services"]["kubeadm"]['controlPlaneEndpoint'],
                'token': join_dict['token'],
                'caCertHashes': [
                    join_dict['discovery-token-ca-cert-hash']
                ]
            }
        },
        'controlPlane': {
            'certificateKey': join_dict['certificate-key'],
            'localAPIEndpoint': {
                'advertiseAddress': node_config['internal_address'],
            }
        }
    }

    # TODO: when k8s v1.21 is excluded from Kubemarine, patches should be added to InitConfiguration unconditionally
    if "v1.21" not in cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]:
        join_config['patches'] = {'directory': '/etc/kubernetes/patches'}

    if cluster.inventory['services']['kubeadm']['controllerManager']['extraArgs'].get(
            'external-cloud-volume-plugin'):
        join_config['nodeRegistration'] = {
            'kubeletExtraArgs': {
                'cloud-provider': 'external'
            }
        }

    if 'worker' in node_config['roles']:
        join_config.setdefault('nodeRegistration', {})['taints'] = []

    configure_container_runtime(cluster, join_config)

    config = get_kubeadm_config(cluster.inventory) + "---\n" + yaml.dump(join_config, default_flow_style=False)

    utils.dump_file(cluster, config, 'join-config_%s.yaml' % node_name)

    log.debug("Uploading init config to control-plane '%s'..." % node_name)
    node.sudo("mkdir -p /etc/kubernetes")
    node.put(io.StringIO(config), '/etc/kubernetes/join-config.yaml', sudo=True)

    # put control-plane patches
    create_kubeadm_patches_for_node(cluster, node)

    # copy admission config to control-plane
    admission.copy_pss(node)

    # put audit-policy.yaml
    prepare_audit_policy(node)

    # ! ETCD on control-planes can't be initialized in async way, that is why it is necessary to disable async mode !
    log.debug('Joining control-plane \'%s\'...' % node_name)

    node.sudo(
        "kubeadm join "
        " --config=/etc/kubernetes/join-config.yaml "
        " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
        " --v=5",
        hide=False)
    defer.sudo("systemctl restart kubelet")
    copy_admin_config(log, defer)
    defer.flush()

    wait_for_any_pods(cluster, node, apply_filter=node_name)


@contextmanager
def local_admin_config(nodes: NodeGroup) -> Iterator[str]:
    temp_filepath = "/tmp/%s" % uuid.uuid4().hex

    cluster_name = nodes.cluster.inventory['cluster_name']

    try:
        with nodes.new_executor() as exe:
            for defer in exe.group.get_ordered_members_list():
                internal_address = defer.get_config()['internal_address']
                if type(ipaddress.ip_address(internal_address)) is ipaddress.IPv6Address:
                    internal_address = f"[{internal_address}]"

                defer.sudo(
                    f"cp /root/.kube/config {temp_filepath} "
                    f"&& sudo sed -i 's/{cluster_name}/{internal_address}/' {temp_filepath}")
        yield temp_filepath
    finally:
        nodes.sudo(f'rm -f {temp_filepath}')


def copy_admin_config(logger: log.EnhancedLogger, nodes: AbstractGroup[RunResult]) -> None:
    logger.debug("Setting up admin-config...")
    command = "mkdir -p /root/.kube && sudo cp -f /etc/kubernetes/admin.conf /root/.kube/config"
    nodes.sudo(command)


def fetch_admin_config(cluster: KubernetesCluster) -> str:
    log = cluster.log

    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    log.debug(f"Downloading kubeconfig from node {first_control_plane.get_node_name()!r}...")

    kubeconfig = list(first_control_plane.sudo('cat /root/.kube/config').values())[0].stdout

    # Replace cluster FQDN with ip
    public_cluster_ip = cluster.inventory.get('public_cluster_ip')
    if public_cluster_ip:
        if type(ipaddress.ip_address(public_cluster_ip)) is ipaddress.IPv6Address:
            public_cluster_ip = f"[{public_cluster_ip}]"
        cluster_name = cluster.inventory['cluster_name']
        kubeconfig = kubeconfig.replace(cluster_name, public_cluster_ip)

    kubeconfig_filename = os.path.abspath("kubeconfig")
    utils.dump_file(cluster.context, kubeconfig, kubeconfig_filename, dump_location=False)
    cluster.log.debug(f"Kubeconfig saved to {kubeconfig_filename}")

    return kubeconfig_filename


def get_join_dict(group: NodeGroup) -> dict:
    cluster: KubernetesCluster = group.cluster
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    token_result = first_control_plane.sudo("kubeadm token create --print-join-command", hide=False)
    join_strings = list(token_result.values())[0].stdout.rstrip("\n")

    join_dict = {"worker_join_command": join_strings}
    join_array = join_strings[join_strings.find("--"):].split()
    for idx, _ in enumerate(join_array):
        current_string = join_array[idx]
        if "--" in current_string:
            join_dict[current_string.lstrip("--")] = join_array[idx + 1]

    cert_key_result = first_control_plane.sudo("kubeadm init phase upload-certs --upload-certs")
    cert_key = list(cert_key_result.values())[0].stdout.split("Using certificate key:\n")[1].rstrip("\n")
    join_dict["certificate-key"] = cert_key
    return join_dict


def init_first_control_plane(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    first_control_plane = group.get_first_member()
    node_config = first_control_plane.get_config()
    node_name = first_control_plane.get_node_name()

    init_config: dict = {
        'apiVersion': cluster.inventory["services"]["kubeadm"]['apiVersion'],
        'kind': 'InitConfiguration',
        'localAPIEndpoint': {
            'advertiseAddress': node_config['internal_address']
        }
    }

    # TODO: when k8s v1.21 is excluded from Kubemarine, patches should be added to InitConfiguration unconditionally
    if "v1.21" not in cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]:
        init_config['patches'] = {'directory': '/etc/kubernetes/patches'}

    if cluster.inventory['services']['kubeadm']['controllerManager']['extraArgs'].get(
            'external-cloud-volume-plugin'):
        init_config['nodeRegistration'] = {
            'kubeletExtraArgs': {
                'cloud-provider': 'external'
            }
        }

    if 'worker' in node_config['roles']:
        init_config.setdefault('nodeRegistration', {})['taints'] = []

    configure_container_runtime(cluster, init_config)

    config = get_kubeadm_config(cluster.inventory) + "---\n" + yaml.dump(init_config, default_flow_style=False)

    utils.dump_file(cluster, config, 'init-config_%s.yaml' % node_name)

    log.debug("Uploading init config to initial control_plane...")
    first_control_plane.sudo("mkdir -p /etc/kubernetes")
    first_control_plane.put(io.StringIO(config), '/etc/kubernetes/init-config.yaml', sudo=True)

    # put control-plane patches
    create_kubeadm_patches_for_node(cluster, first_control_plane)

    # copy admission config to first control-plane
    first_control_plane.call(admission.copy_pss)

    # put audit-policy.yaml
    prepare_audit_policy(first_control_plane)

    log.debug("Initializing first control_plane...")
    result = first_control_plane.sudo(
        "kubeadm init"
        " --upload-certs"
        " --config=/etc/kubernetes/init-config.yaml"
        " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
        " --v=5",
        hide=False)

    copy_admin_config(log, first_control_plane)

    kubeconfig_filepath = fetch_admin_config(cluster)
    summary.schedule_report(cluster.context, summary.SummaryItem.KUBECONFIG, kubeconfig_filepath)

    # Invoke method from admission module for applying default PSS or privileged PSP if they are enabled
    first_control_plane.call(admission.apply_admission)

    # Preparing join_dict to init other nodes
    control_plane_lines = list(result.values())[0].stdout. \
                       split("You can now join any number of the control-plane")[1].splitlines()[2:5]
    worker_lines = list(result.values())[0].stdout. \
                       split("Then you can join any number of worker")[1].splitlines()[2:4]
    control_plane_join_command = " ".join([x.replace("\\", "").strip() for x in control_plane_lines])
    worker_join_command = " ".join([x.replace("\\", "").strip() for x in worker_lines])

    # TODO: Get rid of this code and use get_join_dict() method
    args = control_plane_join_command.split("--")
    join_dict = {}
    for arg in args:
        key_val = arg.split(" ")
        if len(key_val) > 1:
            join_dict[key_val[0].strip()] = key_val[1].strip()
    join_dict["worker_join_command"] = worker_join_command
    cluster.context["join_dict"] = join_dict

    wait_for_any_pods(cluster, first_control_plane, apply_filter=node_name)
    # refresh cluster installation status in cluster context
    is_cluster_installed(cluster)


def wait_for_any_pods(cluster: KubernetesCluster, connection: NodeGroup, apply_filter: str = None) -> None:
    plugins.expect_pods(cluster, [
        'kube-apiserver',
        'kube-controller-manager',
        'kube-proxy',
        'kube-scheduler',
        'etcd'
    ], node=connection, apply_filter=apply_filter,
                        timeout=cluster.inventory['globals']['expect']['pods']['kubernetes']['timeout'],
                        retries=cluster.inventory['globals']['expect']['pods']['kubernetes']['retries'])


def wait_uncordon(node: NodeGroup) -> None:
    cluster = node.cluster
    timeout_config = cluster.inventory['globals']['expect']['pods']['kubernetes']
    # This forces to use local API server and waits till it is up.
    with local_admin_config(node) as kubeconfig:
        utils.wait_command_successful(node, f"kubectl --kubeconfig {kubeconfig} uncordon {node.get_node_name()}",
                                      hide=False,
                                      timeout=timeout_config['timeout'],
                                      retries=timeout_config['retries'])


def wait_for_nodes(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    node_names = group.get_nodes_names()

    wait_conditions = {
        "Ready": "True",
        "NetworkUnavailable": "False"
    }
    if len(node_names) > 1:
        status_cmd = "kubectl get nodes %s -o jsonpath='{.items[*].status.conditions[?(@.type==\"%s\")].status}'"
    else:
        status_cmd = "kubectl get nodes %s -o jsonpath='{.status.conditions[?(@.type==\"%s\")].status}'"

    timeout = int(cluster.inventory['globals']['nodes']['ready']['timeout'])
    retries = int(cluster.inventory['globals']['nodes']['ready']['retries'])
    log.debug("Waiting for new kubernetes nodes to become ready, %s retries every %s seconds" % (retries, timeout))
    while retries > 0:
        correct_conditions = 0
        for condition, cond_value in wait_conditions.items():
            result = first_control_plane.sudo(status_cmd % (" ".join(node_names), condition), warn=True)
            node_result = list(result.values())[0]
            if node_result.failed:
                log.debug(f"kubectl exited with non-zero exit code. Haproxy or kube-apiserver are not yet started?")
                log.verbose(node_result)
                break
            condition_results = node_result.stdout.split(" ")
            correct_values = [value for value in condition_results if value == cond_value]
            if len(correct_values) == len(node_names):
                correct_conditions = correct_conditions + 1
                log.debug(f"Condition {condition} is {cond_value} for all nodes.")
            else:
                log.debug(f"Condition {condition} is not met, retrying")
                break

        if correct_conditions == len(wait_conditions):
            log.debug("All nodes are ready!")
            return
        else:
            retries = retries - 1
            time.sleep(timeout)

    raise Exception("Nodes did not become ready in the expected time, %s retries every %s seconds. Try to increase node.ready.retries parameter in globals: https://github.com/Netcracker/KubeMarine/blob/main/documentation/Installation.md#globals" % (retries, timeout))


def init_workers(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    join_dict = cluster.context.get("join_dict", get_join_dict(group))

    join_config = {
        'apiVersion': group.cluster.inventory["services"]["kubeadm"]['apiVersion'],
        'kind': 'JoinConfiguration',
        'discovery': {
            'bootstrapToken': {
                'apiServerEndpoint': cluster.inventory["services"]["kubeadm"]['controlPlaneEndpoint'],
                'token': join_dict['token'],
                'caCertHashes': [
                    join_dict['discovery-token-ca-cert-hash']
                ]
            }
        }
    }

    # TODO: when k8s v1.21 is excluded from Kubemarine, patches should be added to InitConfiguration unconditionally
    if "v1.21" not in cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]:
        join_config['patches'] = {'directory': '/etc/kubernetes/patches'}

    if cluster.inventory['services']['kubeadm']['controllerManager']['extraArgs'].get(
            'external-cloud-volume-plugin'):
        join_config['nodeRegistration'] = {
            'kubeletExtraArgs': {
                'cloud-provider': 'external'
            }
        }

    configure_container_runtime(cluster, join_config)

    config = yaml.dump(join_config, default_flow_style=False)

    utils.dump_file(cluster, config, 'join-config-workers.yaml')

    group.sudo("mkdir -p /etc/kubernetes")
    group.put(io.StringIO(config), '/etc/kubernetes/join-config.yaml', sudo=True)

    # put control-plane patches
    for node in group.get_ordered_members_list():
        create_kubeadm_patches_for_node(cluster, node)

    cluster.log.debug('Joining workers...')

    for node in group.get_ordered_members_list():
        node.sudo(
            "kubeadm join --config=/etc/kubernetes/join-config.yaml"
            " --ignore-preflight-errors='" + cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors'] + "'"
            " --v=5",
            hide=False)


def apply_labels(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    log.debug("Applying additional labels for nodes")
    # TODO: Add "--overwrite-labels" switch
    # TODO: Add labels validation after applying
    control_plane = cluster.nodes["control-plane"].get_first_member()
    with control_plane.new_executor() as exe:
        for node in group.get_ordered_members_configs_list():
            if "labels" not in node:
                log.verbose("No additional labels found for %s" % node['name'])
                continue
            log.verbose("Found additional labels for %s: %s" % (node['name'], node['labels']))
            for key, value in node["labels"].items():
                exe.group.sudo("kubectl label node %s %s=%s" % (node["name"], key, value))

    log.debug("Successfully applied additional labels")

    return control_plane.sudo("kubectl get nodes --show-labels")
    # TODO: Add wait for pods on worker nodes


def apply_taints(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    log.debug("Applying additional taints for nodes")
    control_plane = cluster.nodes["control-plane"].get_first_member()
    with control_plane.new_executor() as exe:
        for node in group.get_ordered_members_configs_list():
            if "taints" not in node:
                log.verbose("No additional taints found for %s" % node['name'])
                continue
            log.verbose("Found additional taints for %s: %s" % (node['name'], node['taints']))
            for taint in node["taints"]:
                exe.group.sudo("kubectl taint node %s %s" % (node["name"], taint))

    log.debug("Successfully applied additional taints")

    return control_plane.sudo(
        "kubectl get nodes -o=jsonpath="
        "'{range .items[*]}{\"node: \"}{.metadata.name}{\"\\ntaints: \"}{.spec.taints}{\"\\n\"}'")


def is_cluster_installed(cluster: KubernetesCluster) -> bool:
    cluster.log.verbose('Searching for already installed cluster...')
    try:
        results = cluster.nodes['control-plane'].sudo('kubectl cluster-info', warn=True, timeout=15)
        for host, result in results.items():
            if 'is running at' in result.stdout:
                cluster.log.verbose('Detected running Kubernetes cluster on %s' % host)
                for line in result.stdout.split("\n"):
                    if 'Kubernetes control plane' in line:
                        cluster.context['controlplain_uri'] = line.split('at ')[1]
                return True
    except Exception as e:
        cluster.log.verbose(e)
    cluster.context['controlplain_uri'] = None
    cluster.log.verbose('Failed to detect any Kubernetes cluster')
    return False


def get_kubeadm_config(inventory: dict) -> str:
    kubeadm_kubelet = yaml.dump(inventory["services"]["kubeadm_kubelet"], default_flow_style=False)
    kubeadm = yaml.dump(inventory["services"]["kubeadm"], default_flow_style=False)
    return f'{kubeadm_kubelet}---\n{kubeadm}'


def upgrade_first_control_plane(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    node_name = first_control_plane.get_node_name()

    if not upgrade_group.has_node(node_name):
        cluster.log.debug("First control-plane \"%s\" upgrade is not required" % node_name)
        return

    cluster.log.debug("Upgrading first control-plane \"%s\"" % node_name)

    # put control-plane patches
    create_kubeadm_patches_for_node(cluster, first_control_plane)

    flags = "-f --certificate-renewal=true --ignore-preflight-errors='%s' --patches=/etc/kubernetes/patches" % cluster.inventory['services']['kubeadm_flags']['ignorePreflightErrors']

    drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
    first_control_plane.sudo(drain_cmd, hide=False)

    upgrade_cri_if_required(first_control_plane)
    fix_flag_kubelet(first_control_plane)

    first_control_plane.sudo(
        f"sudo kubeadm upgrade apply {version} {flags} && "
        f"sudo kubectl uncordon {node_name} && "
        f"sudo systemctl restart kubelet", hide=False)

    copy_admin_config(cluster.log, first_control_plane)

    expect_kubernetes_version(cluster, version, apply_filter=node_name)
    wait_for_any_pods(cluster, first_control_plane, apply_filter=node_name)
    exclude_node_from_upgrade_list(first_control_plane, node_name)


def upgrade_other_control_planes(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()

    for node in cluster.nodes['control-plane'].get_ordered_members_list():
        node_name = node.get_node_name()
        if node_name != first_control_plane.get_node_name():

            if not upgrade_group.has_node(node_name):
                cluster.log.debug("Control-plane \"%s\" upgrade is not required" % node_name)
                continue

            cluster.log.debug("Upgrading control-plane \"%s\"" % node_name)

            # put control-plane patches
            create_kubeadm_patches_for_node(cluster, node)

            drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
            node.sudo(drain_cmd, hide=False)

            upgrade_cri_if_required(node)
            fix_flag_kubelet(node)

            node.sudo(
                f"sudo kubeadm upgrade node --certificate-renewal=true --patches=/etc/kubernetes/patches && "
                f"sudo kubectl uncordon {node_name} && "
                f"sudo systemctl restart kubelet",
                hide=False)

            expect_kubernetes_version(cluster, version, apply_filter=node_name)
            copy_admin_config(cluster.log, node)
            wait_for_any_pods(cluster, node, apply_filter=node_name)
            exclude_node_from_upgrade_list(first_control_plane, node_name)


def upgrade_workers(upgrade_group: NodeGroup, cluster: KubernetesCluster, **drain_kwargs: Any) -> None:
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    first_control_plane = cluster.nodes['control-plane'].get_first_member()

    for node in cluster.nodes['worker'].exclude_group(cluster.nodes['control-plane'])\
            .get_ordered_members_list():
        node_name = node.get_node_name()

        if not upgrade_group.has_node(node_name):
            cluster.log.debug("Worker \"%s\" upgrade is not required" % node_name)
            continue

        cluster.log.debug("Upgrading worker \"%s\"" % node_name)

        # put control-plane patches
        create_kubeadm_patches_for_node(cluster, node)

        drain_cmd = prepare_drain_command(cluster, node_name, **drain_kwargs)
        first_control_plane.sudo(drain_cmd, hide=False)

        upgrade_cri_if_required(node)
        fix_flag_kubelet(node)

        node.sudo(
            "kubeadm upgrade node --certificate-renewal=true --patches=/etc/kubernetes/patches && "
            "sudo systemctl restart kubelet")

        first_control_plane.sudo("kubectl uncordon %s" % node_name, hide=False)

        expect_kubernetes_version(cluster, version, apply_filter=node_name)
        # workers do not have system pods to wait for their start
        exclude_node_from_upgrade_list(first_control_plane, node_name)


def prepare_drain_command(cluster: KubernetesCluster, node_name: str,
                          *,
                          disable_eviction: bool = False,
                          drain_timeout: int = None, grace_period: int = None) -> str:
    drain_globals = static.GLOBALS['nodes']['drain']
    if drain_timeout is None:
        drain_timeout = recalculate_proper_timeout(cluster, drain_globals['timeout'])

    if grace_period is None:
        grace_period = drain_globals['grace_period']

    drain_cmd = f"kubectl drain {node_name} --force --ignore-daemonsets --delete-emptydir-data " \
                f"--timeout={drain_timeout}s --grace-period={grace_period}"
    if disable_eviction:
        drain_cmd += " --disable-eviction=true"
    return drain_cmd


def upgrade_cri_if_required(group: NodeGroup) -> None:
    # currently it is invoked only for single node
    cluster: KubernetesCluster = group.cluster
    log = cluster.log
    cri_impl = cluster.inventory['services']['cri']['containerRuntime']

    if cri_impl in cluster.context["upgrade"]["required"]['packages']:
        cri_packages = cluster.get_package_association_for_node(group.get_host(), cri_impl, 'package_name')

        log.debug(f"Installing {cri_packages} on node: {group.get_node_name()}")
        packages.install(group, include=cri_packages)
        log.debug(f"Restarting all containers on node: {group.get_node_name()}")
        if cri_impl == "docker":
            group.sudo("docker container rm -f $(sudo docker container ls -q)", warn=True)
        else:
            group.sudo("crictl rm -fa", warn=True)
    else:
        log.debug(f"{cri_impl!r} package upgrade is not required")

    # upgrade of sandbox_image is currently not supported for migrate_kubemarine
    if cri_impl == 'containerd' and cluster.context["upgrade"]["required"].get('containerdConfig', False):
        containerd.configure_containerd(group)
    else:
        log.debug(f"{cri_impl!r} configuration upgrade is not required")


def verify_upgrade_versions(cluster: KubernetesCluster) -> None:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    upgrade_version = cluster.context["upgrade_version"]

    k8s_nodes_group = cluster.nodes["worker"].include_group(cluster.nodes['control-plane'])
    for node in k8s_nodes_group.get_ordered_members_list():
        cluster.log.debug(f"Verifying current k8s version for node {node.get_node_name()}")
        result = first_control_plane.sudo("kubectl get nodes "
                                          f"{node.get_node_name()}"
                                          " -o custom-columns='VERSION:.status.nodeInfo.kubeletVersion' "
                                          "| grep -vw ^VERSION ")
        curr_version = list(result.values())[0].stdout
        test_version_upgrade_possible(curr_version, upgrade_version, skip_equal=True)


def get_initial_kubernetes_version(inventory: dict) -> str:
    kubernetes_version: str
    if inventory.get("services", {}).get("kubeadm", {}).get("kubernetesVersion") is not None:
        kubernetes_version = inventory['services']['kubeadm']['kubernetesVersion']
    else:
        kubernetes_version = static.DEFAULTS['services']['kubeadm']['kubernetesVersion']

    return kubernetes_version


def verify_initial_version(inventory: dict, _: KubernetesCluster) -> dict:
    version = get_initial_kubernetes_version(inventory)
    verify_allowed_version(version)
    return inventory


def verify_allowed_version(version: str) -> None:
    allowed_versions = static.KUBERNETES_VERSIONS['compatibility_map'].keys()
    if version not in allowed_versions:
        raise errors.KME('KME0008',
                         version=version,
                         allowed_versions=', '.join(map(repr, allowed_versions)))


def verify_supported_version(target_version: str, logger: log.EnhancedLogger) -> None:
    verify_allowed_version(target_version)
    minor_version = utils.minor_version(target_version)
    supported_versions = static.KUBERNETES_VERSIONS['kubernetes_versions']
    if not supported_versions.get(minor_version, {}).get("supported", False):
        logger.warning(f"Specified target Kubernetes version {target_version!r} - is not supported!")


def expect_kubernetes_version(cluster: KubernetesCluster, version: str,
                              timeout: int = None, retries: int = None,
                              node: NodeGroup = None, apply_filter: str = None) -> None:
    if timeout is None:
        timeout = cluster.globals['nodes']['expect']['kubernetes_version']['timeout']
    if retries is None:
        retries = cluster.globals['nodes']['expect']['kubernetes_version']['retries']

    cluster.log.debug("Expecting Kubernetes version %s" % version)
    cluster.log.debug("Max expectation time: %ss" % (timeout * retries))

    cluster.log.debug("Waiting for nodes...\n")

    if node is None:
        node = cluster.nodes['control-plane'].get_first_member()

    command = 'kubectl get nodes -o=wide'
    if apply_filter is not None:
        command += ' | grep -w %s' % apply_filter

    while retries > 0:
        result = node.sudo(command, warn=True)
        stdout = list(result.values())[0].stdout
        cluster.log.verbose(stdout)
        nodes_version_correct = True
        for stdout_line in iter(stdout.splitlines()):
            if version not in stdout_line:
                nodes_version_correct = False
                cluster.log.verbose("Invalid version detected: %s\n" % stdout_line)

        if nodes_version_correct:
            cluster.log.debug("Nodes have correct Kubernetes version = %s" % version)
            cluster.log.debug(result)
            return
        else:
            retries -= 1
            cluster.log.debug("Some nodes have invalid Kubernetes version... (%ss left)" % (retries * timeout), result)
            time.sleep(timeout)

    raise Exception('In the expected time, the nodes did not receive correct Kubernetes version')


def test_version_upgrade_possible(old: str, new: str, skip_equal: bool = False) -> None:
    versions_unchanged = {
        'old': old.strip(),
        'new': new.strip()
    }
    versions: Dict[str, Tuple[int, int, int]] = {}

    for v_type, version in versions_unchanged.items():
        versions[v_type] = utils.version_key(version)

    # test new is greater than old
    if versions['old'] > versions['new']:
        raise Exception(ERROR_DOWNGRADE % (versions_unchanged['old'], versions_unchanged['new']))

    # test new is the same as old
    if versions['old'] == versions['new'] and not skip_equal:
        raise Exception(ERROR_SAME % (versions_unchanged['old'], versions_unchanged['new']))

    # test major step is not greater than 1
    if versions['new'][0] - versions['old'][0] > 1:
        raise Exception(ERROR_MAJOR_RANGE_EXCEEDED % (versions_unchanged['old'], versions_unchanged['new']))

    # test minor step is not greater than 1
    if versions['new'][1] - versions['old'][1] > 1:
        raise Exception(ERROR_MINOR_RANGE_EXCEEDED % (versions_unchanged['old'], versions_unchanged['new']))


def recalculate_proper_timeout(cluster: KubernetesCluster, timeout: int) -> int:
    try:
        amount_str = cluster.nodes['control-plane'].get_first_member().sudo('kubectl get pods -A | wc -l').get_simple_out()
        return timeout * int(amount_str)
    except Exception:
        return timeout * 10 * cluster.nodes['all'].nodes_amount()


def configure_container_runtime(cluster: KubernetesCluster, kubeadm_config: dict) -> None:
    if cluster.inventory['services']['cri']['containerRuntime'] == "containerd":
        if 'nodeRegistration' not in kubeadm_config:
            kubeadm_config['nodeRegistration'] = {}
        if 'kubeletExtraArgs' not in kubeadm_config['nodeRegistration']:
            kubeadm_config['nodeRegistration']['kubeletExtraArgs'] = {}

        kubeadm_config['nodeRegistration']['criSocket'] = '/var/run/containerd/containerd.sock'

        minor_version = int(cluster.inventory["services"]["kubeadm"]["kubernetesVersion"].split('.')[1])
        if minor_version < 27:
            kubeadm_config['nodeRegistration']['kubeletExtraArgs']['container-runtime'] = 'remote'

        kubeadm_config['nodeRegistration']['kubeletExtraArgs']['container-runtime-endpoint'] = \
            'unix:///run/containerd/containerd.sock'


def exclude_node_from_upgrade_list(first_control_plane: NodeGroup, node_name: str) -> None:
    first_control_plane.sudo('sed -i \'/%s/d\' /etc/kubernetes/nodes-k8s-versions.txt' % node_name, warn=True)


def autodetect_non_upgraded_nodes(cluster: KubernetesCluster, future_version: str) -> List[str]:
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    try:
        nodes_list_result = first_control_plane.sudo('[ ! -f /etc/kubernetes/nodes-k8s-versions.txt ] && '
                                              'sudo kubectl get nodes -o custom-columns=\''
                                              'VERSION:.status.nodeInfo.kubeletVersion,'
                                              'NAME:.metadata.name,'
                                              'STATUS:.status.conditions[-1].type\' '
                                              '| sed -n \'1!p\' | tr -s \' \' '
                                              '| sed \'1 i\\# This file contains a cached list of nodes and versions '
                                              'required to continue the Kubernetes upgrade procedure if it fails. '
                                              'If all the nodes are completely updated or you manually fixed the '
                                              'problem that occurred during the upgrade, you can delete it.\' '
                                              '| sudo tee /etc/kubernetes/nodes-k8s-versions.txt; '
                                              'sudo cat /etc/kubernetes/nodes-k8s-versions.txt') \
            .get_simple_out()
        cluster.log.verbose("Remote response with nodes description:\n%s" % nodes_list_result)
    except Exception as e:
        cluster.log.warning("Failed to detect cluster status before upgrade. All nodes will be scheduled for upgrade.")
        cluster.log.verbose(e)
        return cluster.nodes['all'].get_nodes_names()

    detected_nodes_lines = nodes_list_result.splitlines()

    if not detected_nodes_lines:
        raise Exception('Remote result did not returned any lines containing node info')

    upgrade_list = []
    for line in detected_nodes_lines:
        line = line.strip()

        # comes from nodes-k8s-versions.txt content as a comment symbol
        if line[0] == '#':
            continue
        version, node_name, status = line.split(' ')
        if version != future_version:
            cluster.log.verbose("Node \"%s\" has version \"%s\" and scheduled for upgrade." % (node_name, version))
            upgrade_list.append(node_name)
        elif status != 'Ready':
            cluster.log.verbose("Node \"%s\" is not ready and scheduled for upgrade." % node_name)
            upgrade_list.append(node_name)
        else:
            cluster.log.verbose("Node \"%s\" already upgraded." % node_name)

    return upgrade_list


def get_group_for_upgrade(cluster: KubernetesCluster) -> NodeGroup:
    upgrade_group: Optional[NodeGroup] = cluster.context.get('upgrade_group')
    if upgrade_group is not None:
        return upgrade_group

    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    if cluster.procedure_inventory.get('upgrade_nodes'):
        nodes_for_upgrade = []
        for node in cluster.procedure_inventory['upgrade_nodes']:
            if isinstance(node, str):
                node_name = node
            else:
                node_name = node['name']
            nodes_for_upgrade.append(node_name)
            cluster.log.verbose("Node \"%s\" manually scheduled for upgrade." % node_name)
            cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt', warn=True)
    else:
        nodes_for_upgrade = autodetect_non_upgraded_nodes(cluster, version)

    upgrade_group = cluster.make_group_from_nodes(nodes_for_upgrade)
    cluster.context['upgrade_group'] = upgrade_group

    return upgrade_group


def images_grouped_prepull(group: NodeGroup, group_size: int = None) -> RunnersGroupResult:
    """
    Prepull kubeadm images on group, separated on sub-groups with certain group size. Separation required to avoid high
    load on images repository server, when using large clusters.
    :param group: NodeGroup where prepull should be performed.
    :param group_size: integer number of nodes per group. Will be automatically used from procedure_yaml or globals, if not set.
    :return: String results from all nodes in presented group.
    """

    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    if group_size is None:
        group_size = cluster.procedure_inventory.get('prepull_group_size')

    if group_size is None:
        log.verbose("Group size is not set in procedure inventory, a default one will be used")
        group_size = cluster.globals['prepull_group_size']

    nodes_amount = group.nodes_amount()

    # group_size should be greater than 0
    if nodes_amount != 0 and nodes_amount < group_size:
        group_size = nodes_amount

    groups_amount = math.ceil(nodes_amount / group_size)

    log.verbose("Nodes amount: %s\nGroup size: %s\nGroups amount: %s" % (nodes_amount, group_size, groups_amount))
    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        nodes = exe.group.get_ordered_members_list()
        for group_i in range(groups_amount):
            log.verbose('Prepulling images for group #%s...' % group_i)
            # RemoteExecutor used for future cases, when some nodes will require another/additional actions for prepull
            for node_i in range(group_i*group_size, (group_i*group_size)+group_size):
                if node_i < nodes_amount:
                    images_prepull(nodes[node_i], collector=collector)

    return collector.result


def images_prepull(group: DeferredGroup, collector: CollectorCallback) -> Token:
    """
    Prepull kubeadm images on group.

    :param group: NodeGroup where prepull should be performed.
    :param collector: CollectorCallback instance
    :return: NodeGroupResult from all nodes in presented group.
    """

    config = get_kubeadm_config(group.cluster.inventory)
    kubeadm_init: dict = {
        'apiVersion': group.cluster.inventory["services"]["kubeadm"]['apiVersion'],
        'kind': 'InitConfiguration',
    }

    configure_container_runtime(group.cluster, kubeadm_init)
    config = f'{config}---\n{yaml.dump(kubeadm_init, default_flow_style=False)}'

    group.put(io.StringIO(config), '/etc/kubernetes/prepull-config.yaml', sudo=True)

    return group.sudo("kubeadm config images pull --config=/etc/kubernetes/prepull-config.yaml",
                      callback=collector)


def schedule_running_nodes_report(cluster: KubernetesCluster) -> None:
    summary.schedule_delayed_report(cluster, exec_running_nodes_report)


def exec_running_nodes_report(cluster: KubernetesCluster) -> None:
    nodes_description = get_nodes_description(cluster)
    actual_roles = get_actual_roles(nodes_description)
    nodes_conditions = get_nodes_conditions(nodes_description)
    nodes_names = actual_roles.keys()
    for role in ('control-plane', 'worker'):
        members = 0
        ready = 0
        for name in nodes_names:
            if role in actual_roles[name]:
                members += 1
                conditions = nodes_conditions[name]
                if conditions.get('Ready', {}).get('status') == 'True' \
                        and conditions.get('NetworkUnavailable', {}).get('status') == 'False':
                    ready += 1

        property = summary.SummaryItem.CONTROL_PLANES if role == 'control-plane' else summary.SummaryItem.WORKERS
        value = f'{ready}/{members}'
        summary.schedule_report(cluster.context, property, value)


def get_nodes_description_cmd() -> str:
    return 'kubectl get node -o yaml'


def get_nodes_description(cluster: KubernetesCluster) -> dict:
    cmd = get_nodes_description_cmd()
    result = cluster.nodes['control-plane'].get_final_nodes().get_any_member().sudo(cmd)
    cluster.log.verbose(result)
    data: dict = yaml.safe_load(list(result.values())[0].stdout)
    return data


def get_actual_roles(nodes_description: dict) -> Dict[str, List[str]]:
    result: Dict[str, List[str]] = {}
    for node_description in nodes_description['items']:
        node_name = node_description['metadata']['name']
        labels = node_description['metadata']['labels']
        result[node_name] = []
        # TODO check label accordingly to Kubernetes version
        if 'node-role.kubernetes.io/master' in labels or 'node-role.kubernetes.io/control-plane' in labels:
            result[node_name].append('control-plane')
        if 'node-role.kubernetes.io/worker' in labels:
            result[node_name].append('worker')

    return result


def get_nodes_conditions(nodes_description: dict) -> Dict[str, Dict[str, dict]]:
    result = {}
    for node_description in nodes_description['items']:
        node_name = node_description['metadata']['name']
        conditions_by_type: Dict[str, dict] = {}
        result[node_name] = conditions_by_type
        for condition in node_description['status']['conditions']:
            conditions_by_type[condition['type']] = condition

    return result


# function to get dictionary of flags to be patched for a given control plane item and a given node
def get_patched_flags_for_control_plane_item(inventory: dict, control_plane_item: str, node: NodeConfig) -> Dict[str, str]:
    flags = {}

    for n in inventory['services']['kubeadm_patches'][control_plane_item]:
        if n.get('groups') is not None and list(set(node['roles']) & set(n['groups'])):
            if n.get('patch') is not None:
                for arg, value in n['patch'].items():
                    flags[arg] = value
        if n.get('nodes') is not None and node['name'] in n['nodes']:
            if n.get('patch') is not None:
                for arg, value in n['patch'].items():
                    flags[arg] = value

    # we always set binding-address to the node's internal address for apiServer
    if control_plane_item == 'apiServer' and 'control-plane' in node['roles']:
        flags['bind-address'] = node['internal_address']

    return flags


# function to create kubeadm patches and put them to a node
def create_kubeadm_patches_for_node(cluster: KubernetesCluster, node: NodeGroup) -> None:
    cluster.log.verbose(f"Create and upload kubeadm patches to %s..." % node.get_node_name())
    node.sudo('sudo rm -rf /etc/kubernetes/patches ; sudo mkdir -p /etc/kubernetes/patches', warn=True)

    control_plane_patch_files = {
        'apiServer' : 'kube-apiserver+json.json',
        'etcd' : 'etcd+json.json',
        'controllerManager' : 'kube-controller-manager+json.json',
        'scheduler' : 'kube-scheduler_json.json',
        'kubelet' : 'kubeletconfiguration.yaml'
    }

    # read patches content from inventory and upload patch files to a node
    node_config = node.get_config()
    for control_plane_item in cluster.inventory['services']['kubeadm_patches']:
        patched_flags = get_patched_flags_for_control_plane_item(cluster.inventory, control_plane_item, node_config)
        if patched_flags:
            if control_plane_item == 'kubelet':
                template_filename = 'templates/patches/kubelet.yaml.j2'
            else:
                template_filename = 'templates/patches/control-plane-pod.json.j2'

            control_plane_patch = Template(utils.read_internal(template_filename)).render(flags=patched_flags)
            patch_file = '/etc/kubernetes/patches/' + control_plane_patch_files[control_plane_item]
            node.put(io.StringIO(control_plane_patch + "\n"), patch_file, sudo=True)
            node.sudo(f'chmod 644 {patch_file}')


def fix_flag_kubelet(group: NodeGroup) -> bool:
    kubeadm_flags_file = "/var/lib/kubelet/kubeadm-flags.env"
    cluster = group.cluster
    version = cluster.inventory["services"]["kubeadm"]["kubernetesVersion"]
    sandbox_image = containerd.get_sandbox_image(cluster.inventory['services']['cri'])
    infra_image_flag = f"--pod-infra-container-image={sandbox_image}"
    container_runtime_flag = '--container-runtime=remote'

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            node.sudo(f"cat {kubeadm_flags_file}", callback=collector)

    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            kubeadm_flags = collector.result[node.get_host()].stdout
            updated_kubeadm_flags = kubeadm_flags
            if utils.version_key(version) >= utils.version_key("v1.27.0"):
                # remove the deprecated kubelet flag for versions starting from 1.27.0
                updated_kubeadm_flags = updated_kubeadm_flags.replace(container_runtime_flag, '')

            if infra_image_flag not in updated_kubeadm_flags:
                # patch --pod-infra-container-image with target sandbox_image
                updated_kubeadm_flags = config_changer(updated_kubeadm_flags, infra_image_flag)

            if kubeadm_flags != updated_kubeadm_flags:
                cluster.log.debug(f"Patching {kubeadm_flags_file} on {node.get_node_name()} node...")
                node.put(io.StringIO(updated_kubeadm_flags), kubeadm_flags_file, backup=True, sudo=True)

    # If file is changed on at least one node, last results will be not empty
    return len(exe.get_last_results()) > 0


def config_changer(config: str, word: str) -> str:
    equal_pos = word.find("=") + 1
    param_begin_pos = config.find(word[:equal_pos])
    if param_begin_pos != -1:
        param_end_pos = config[param_begin_pos:].find(" ")
        if param_end_pos == -1:
            return config[:param_begin_pos] + word + "\""
        return config[:param_begin_pos] + word + config[param_end_pos + param_begin_pos:]
    else:
        param_end_pos = config.rfind("\"")
        return config[:param_end_pos] + " " + word[:] + "\""


def prepare_audit_policy(group: NodeGroup) -> None:
    """
    Prepare audit-policy.yaml and all necessary directories.
    """
    cluster: KubernetesCluster = group.cluster
    api_server_extra_args = cluster.inventory['services']['kubeadm']['apiServer']['extraArgs']
    audit_log_dir = os.path.dirname(api_server_extra_args['audit-log-path'])
    audit_file_name = api_server_extra_args['audit-policy-file']
    audit_policy_dir = os.path.dirname(audit_file_name)
    group.sudo(f"mkdir -p {audit_log_dir} && sudo mkdir -p {audit_policy_dir}")

    cluster.log.debug("Configure audit cluster policy")
    policy_config = cluster.inventory['services']['audit']['cluster_policy']
    policy_config_file = yaml.dump(policy_config)
    utils.dump_file(cluster, policy_config_file, 'audit-policy.yaml')
    # upload rules on cluster
    group.put(io.StringIO(policy_config_file), audit_file_name, sudo=True, backup=True)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from importlib import import_module
from copy import deepcopy
from typing import Optional, Dict, Any, Tuple, List

import yaml

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.errors import KME
from kubemarine import jinja, keepalived, haproxy
from kubemarine.core import utils, static, log, os
from kubemarine.core.yaml_merger import default_merger
from kubemarine.cri.containerd import contains_old_format_properties

# All enrichment procedures should not connect to any node.
# The information about nodes should be collected within KubernetesCluster#detect_nodes_context().
DEFAULT_ENRICHMENT_FNS = [
    "kubemarine.core.schema.verify_inventory",
    "kubemarine.core.defaults.merge_defaults",
    "kubemarine.kubernetes.verify_initial_version",
    "kubemarine.admission.enrich_default_admission",
    "kubemarine.kubernetes.add_node_enrichment",
    "kubemarine.core.defaults.calculate_node_names",
    "kubemarine.kubernetes.remove_node_enrichment",
    "kubemarine.controlplane.controlplane_node_enrichment",
    "kubemarine.core.defaults.append_controlplain",
    "kubemarine.kubernetes.enrich_upgrade_inventory",
    "kubemarine.kubernetes.enrich_restore_inventory",
    "kubemarine.core.defaults.compile_inventory",
    "kubemarine.core.defaults.manage_true_false_values",
    "kubemarine.plugins.enrich_upgrade_inventory",
    "kubemarine.packages.enrich_inventory",
    "kubemarine.packages.enrich_upgrade_inventory",
    "kubemarine.packages.enrich_migrate_cri_inventory",
    "kubemarine.packages.enrich_inventory_apply_defaults",
    "kubemarine.thirdparties.enrich_upgrade_inventory",
    "kubemarine.thirdparties.enrich_restore_inventory",
    "kubemarine.thirdparties.enrich_migrate_cri_inventory",
    "kubemarine.admission.manage_enrichment",
    "kubemarine.cri.containerd.enrich_migrate_cri_inventory",
    "kubemarine.core.defaults.apply_registry",
    "kubemarine.cri.enrich_upgrade_inventory",
    "kubemarine.core.defaults.verify_node_names",
    "kubemarine.core.defaults.apply_defaults",
    "kubemarine.keepalived.enrich_inventory_apply_defaults",
    "kubemarine.haproxy.enrich_inventory",
    "kubemarine.kubernetes.enrich_inventory",
    "kubemarine.admission.enrich_inventory",
    "kubemarine.kubernetes_accounts.enrich_inventory",
    "kubemarine.plugins.calico.enrich_inventory",
    "kubemarine.plugins.nginx_ingress.cert_renew_enrichment",
    "kubemarine.plugins.nginx_ingress.enrich_inventory",
    "kubemarine.plugins.local_path_provisioner.enrich_inventory",
    "kubemarine.plugins.kubernetes_dashboard.enrich_inventory",
    "kubemarine.core.defaults.calculate_nodegroups",
    "kubemarine.keepalived.enrich_inventory_calculate_nodegroup",
    "kubemarine.thirdparties.enrich_inventory_apply_defaults",
    "kubemarine.system.verify_inventory",
    "kubemarine.system.enrich_etc_hosts",
    "kubemarine.packages.enrich_inventory_include_all",
    "kubemarine.audit.verify_inventory",
    "kubemarine.plugins.enrich_inventory",
    "kubemarine.plugins.verify_inventory",
    "kubemarine.plugins.builtin.verify_inventory",
    "kubemarine.k8s_certs.renew_verify",
    "kubemarine.cri.enrich_inventory",
    "kubemarine.system.enrich_kernel_modules"
]

supported_defaults = {
    'rbac': {
        'account_defaults': 'accounts'
    },
    'node_defaults': 'nodes',
}

invalid_node_name_regex = re.compile("[^a-z-.\\d]", re.M)
escaped_expression_regex = re.compile('({%[\\s*|]raw[\\s*|]%}.*?{%[\\s*|]endraw[\\s*|]%})', re.M)
jinja_query_regex = re.compile("{{ .* }}", re.M)


def apply_defaults(inventory: dict, cluster: KubernetesCluster) -> dict:
    recursive_apply_defaults(supported_defaults, inventory)

    for i, node in enumerate(inventory["nodes"]):
        address = cluster.get_access_address_from_node(node)

        # we definitely know how to connect
        cluster.inventory["nodes"][i]["connect_to"] = address

        if not cluster.context["nodes"].get(address):
            cluster.context["nodes"][address] = {}

        if address not in cluster.ips["all"]:
            cluster.ips['all'].append(address)

        for role in node.get("roles"):
            if role not in cluster.roles:
                cluster.roles.append(role)
                cluster.ips[role] = []
            if address not in cluster.ips[role]:
                cluster.ips[role].append(address)

    return inventory


def apply_registry(inventory: dict, cluster: KubernetesCluster) -> dict:

    if not inventory.get('registry'):
        cluster.log.verbose('Unified registry is not used')
        return inventory

    thirdparties_address = None
    containerd_endpoints = None
    protocol = None

    # registry contains either 'endpoints' or 'address' that is validated by JSON schema.
    if inventory['registry'].get('endpoints'):
        registry_mirror_address, containerd_endpoints, thirdparties_address = apply_registry_endpoints(inventory)
    else:
        if inventory['registry'].get('docker_port'):
            registry_mirror_address = "%s:%s" % (inventory['registry']['address'], inventory['registry']['docker_port'])
        else:
            registry_mirror_address = inventory['registry']['address']

        protocol = 'http'
        if inventory['registry'].get('ssl', False):
            protocol = 'https'

        if inventory['registry'].get('webserver', False):
            thirdparties_address = f"{protocol}://{inventory['registry']['address']}"

    # Patch kubeadm imageRepository and plugin_defaults registry
    if cluster.raw_inventory.get('services', {}).get('kubeadm', {}).get('imageRepository') is None:
        inventory['services']['kubeadm']["imageRepository"] = registry_mirror_address
    inventory['plugin_defaults']['installation'].setdefault('registry', registry_mirror_address)

    cri_impl = inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker" and inventory['registry'].get('endpoints'):
        raise KME('KME0007')

    if cri_impl == "docker":

        if protocol == 'http':
            if inventory['services']['cri']['dockerConfig'].get("insecure-registries") is None:
                inventory['services']['cri']['dockerConfig']["insecure-registries"] = []
            insecure_registries = inventory['services']['cri']['dockerConfig']["insecure-registries"]
            insecure_registries.append(registry_mirror_address)
            inventory['services']['cri']['dockerConfig']["insecure-registries"] = list(set(insecure_registries))

        if inventory['services']['cri']['dockerConfig'].get("registry-mirrors") is None:
            inventory['services']['cri']['dockerConfig']["registry-mirrors"] = []

        registry_mirrors = inventory['services']['cri']['dockerConfig']["registry-mirrors"]
        registry_mirrors.append(f"{protocol}://{registry_mirror_address}")
        inventory['services']['cri']['dockerConfig']["registry-mirrors"] = list(set(registry_mirrors))

    elif cri_impl == "containerd":
        if not containerd_endpoints:
            containerd_endpoints = ["%s://%s" % (protocol, registry_mirror_address)]

        old_format_result, _ = contains_old_format_properties(inventory)
        # todo remove p3_reconfigure_registries after next release
        if old_format_result or not cluster.context.get('p3_reconfigure_registries', True):
            # Add registry info in old format
            registry_section = f'plugins."io.containerd.grpc.v1.cri".registry.mirrors."{registry_mirror_address}"'
            containerd_config = inventory['services']['cri']['containerdConfig']
            if not containerd_config.get(registry_section):
                containerd_config[registry_section] = {
                    'endpoint': containerd_endpoints
                }
        else:
            # Add registry info in new format
            old_registry_config = inventory['services']['cri'].get('containerdRegistriesConfig', {})
            inventory['services']['cri']['containerdRegistriesConfig'] = {registry_mirror_address: {
                                     f'host."{endpoint}"': {'capabilities': ['pull', 'resolve']}
                                     for endpoint in containerd_endpoints
                                 }}
            default_merger.merge(inventory['services']['cri']['containerdRegistriesConfig'], old_registry_config)

    from kubemarine import thirdparties

    if thirdparties_address:
        for destination, config in inventory['services']['thirdparties'].items():
            if not thirdparties.is_default_thirdparty(destination) or isinstance(config, str) or 'source' in config:
                continue

            source, sha1 = thirdparties.get_default_thirdparty_identity(cluster.inventory, destination, in_public=False)
            source = source.format(registry=thirdparties_address)
            config['source'] = source
            if 'sha1' not in config:
                config['sha1'] = sha1

    return inventory


def apply_registry_endpoints(inventory: dict) -> Tuple[str, List[str], Optional[str]]:

    if not inventory['registry'].get('mirror_registry'):
        inventory['registry']['mirror_registry'] = 'registry.cluster.local'

    registry_mirror_address = inventory['registry']['mirror_registry']

    # todo Currently registry.endpoints is used only for containerd registry mirrors, but it can be provided explicitly.
    #  Probably we could make endpoints optional in this case.
    containerd_endpoints = inventory['registry']['endpoints']
    thirdparties_address = inventory['registry'].get('thirdparties')

    return registry_mirror_address, containerd_endpoints, thirdparties_address


def append_controlplain(inventory: dict, cluster: Optional[KubernetesCluster]) -> dict:

    if inventory.get('control_plain', {}).get('internal') and inventory.get('control_plain', {}).get('external'):
        if cluster:
            cluster.log.verbose('Control plains are set manually, nothing to detect.')
        return inventory

    if cluster:
        cluster.log.verbose('Detecting control plains...')

    # calculate controlplain ips
    internal_address: Optional[str] = None
    internal_address_source: Optional[str] = None
    external_address: Optional[str] = None
    external_address_source: Optional[str] = None

    balancer_names = keepalived.get_all_balancer_names(inventory, final=True)
    # vrrp_ip section is not enriched yet
    # If no VRRP IPs or no balancers are configured, Keepalived is not enabled.
    if inventory.get('vrrp_ips') and balancer_names:
        for i, item in enumerate(inventory['vrrp_ips']):
            if isinstance(item, str):
                if internal_address is None:
                    internal_address = item
                    internal_address_source = 'vrrp_ip[%s]' % i
            else:
                if haproxy.is_vrrp_not_bind(item):
                    continue
                final_hosts = item.get('hosts', balancer_names)
                # There is a small gap here.
                # The check is invoked when inventory is not yet compiled, so checking names for equality is not fair.
                if not any((host['name'] if isinstance(host, dict) else host) in balancer_names
                           for host in final_hosts):
                    continue
                if internal_address is None or item.get('control_endpoint', False):
                    internal_address = item['ip']
                    internal_address_source = 'vrrp_ip[%s]' % i
                if item.get('floating_ip') and (external_address is None or item.get('control_endpoint', False)):
                    external_address = item['floating_ip']
                    external_address_source = 'vrrp_ip[%s]' % i

    if internal_address is not None and external_address is None and cluster:
        cluster.log.warning('VRRP_IPs has an internal address, but do not have an external one. Your configuration may be incorrect. Trying to handle this problem automatically...')

    if internal_address is None or external_address is None:
        # 'master' role is not deleted due to unit tests are not refactored
        for role in ['balancer', 'control-plane', 'master']:
            # nodes are not compiled to groups yet
            for node in inventory['nodes']:
                if role in node['roles'] and 'remove_node' not in node['roles']:
                    if internal_address is None or node.get('control_endpoint', False):
                        internal_address = node['internal_address']
                        internal_address_source = f"{role} \"{node['name']}\""
                    if node.get('address') and (external_address is None or node.get('control_endpoint', False)):
                        external_address = node['address']
                        external_address_source = f"{role} \"{node['name']}\""

    if external_address is None:
        if cluster:
            cluster.log.warning('Failed to detect external control plain. Something may work incorrect!')
        external_address = internal_address

    if cluster:
        cluster.log.debug('Control plains:\n   Internal: %s (%s)\n   External: %s (%s)' % (internal_address, internal_address_source, external_address, external_address_source))

    # apply controlplain ips
    if not inventory.get('control_plain'):
        inventory['control_plain'] = {}

    if not inventory['control_plain'].get('internal'):
        inventory['control_plain']['internal'] = internal_address

    if not inventory['control_plain'].get('external'):
        inventory['control_plain']['external'] = external_address

    return inventory


def recursive_apply_defaults(defaults: dict, section: dict) -> None:
    for key, value in defaults.items():
        if isinstance(value, dict) and section.get(key) is not None and section[key]:
            recursive_apply_defaults(value, section[key])
        # check if target section exists and not empty
        elif section.get(value) is not None:
            for i, custom_value in enumerate(section[value]):
                # copy defaults as new dict, to avoid problems with memory links
                default_value = deepcopy(section[key])

                # update defaults with custom-defined node configs
                # TODO: Use deepmerge instead of update
                default_value.update(custom_value)

                # replace old node config with merged one
                section[value][i] = default_value


def calculate_node_names(inventory: dict, cluster: KubernetesCluster) -> dict:
    roles_iterators: Dict[str, int] = {}
    for i, node in enumerate(inventory['nodes']):
        # 'master' role is not deleted because calculate_node_names() can be run over initial inventory,
        # that still supports the old role.
        for role_name in ['control-plane', 'master', 'worker', 'balancer']:
            if role_name in node['roles']:
                # The idea is this:
                # If the name is already specified, we must skip this node,
                # however, we must consider that we already have a node of this type
                # and increase this type iterator
                # As a result, we get such an algorithm. For example, with the following inventory:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                # We should get the following calculation result:
                #
                # - name: k8s-control-plane-1, roles: ['control-plane']
                # - name: control-plane-2, roles: ['control-plane']
                # - name: k8s-control-plane-3, roles: ['control-plane']
                #
                role_i = roles_iterators.get(role_name, 1)
                roles_iterators[role_name] = role_i + 1
                if node.get('name') is None:
                    if role_name == 'master':
                        role_name = 'control-plane'
                    new_name = '%s-%s' % (role_name, role_i)
                    cluster.log.debug(f"Assigning name {new_name} to node {cluster.get_access_address_from_node(node)}")
                    node['name'] = new_name
    return inventory


def verify_node_names(inventory: dict, _: KubernetesCluster) -> dict:
    known_names = []
    for i, node in enumerate(inventory['nodes']):
        node_name = node['name']
        if node_name in known_names:
            raise Exception('Node name %s is duplicated in configfile' % node_name)
        if re.findall(invalid_node_name_regex, node_name):
            raise Exception('Node name \"%s\" contains invalid characters. A DNS-1123 subdomain must consist of lower '
                            'case alphanumeric characters, \'-\' or \'.\'' % node_name)
        known_names.append(node_name)
    return inventory


def calculate_nodegroups(inventory: dict, cluster: KubernetesCluster) -> dict:
    for role in cluster.ips.keys():
        cluster.nodes[role] = cluster.make_group(cluster.ips[role])
    return inventory


def merge_defaults(inventory: dict, cluster: KubernetesCluster) -> dict:
    base_inventory = deepcopy(static.DEFAULTS)

    inventory = default_merger.merge(base_inventory, inventory)
    # it is necessary to temporary put half-compiled inventory to cluster inventory field
    cluster._inventory = inventory
    return inventory


def enrich_inventory(cluster: KubernetesCluster, inventory: dict,
                     make_dumps: bool = True, enrichment_functions: List[str] = None) -> dict:
    if not enrichment_functions:
        enrichment_functions = DEFAULT_ENRICHMENT_FNS

    # run required fields calculation
    for enrichment_fn in enrichment_functions:
        fn_package_name, fn_method_name = enrichment_fn.rsplit('.', 1)
        mod = import_module(fn_package_name)
        cluster.log.verbose('Calling fn "%s"' % enrichment_fn)
        inventory = getattr(mod, fn_method_name)(inventory, cluster)

    cluster.log.verbose('Enrichment finished!')

    if make_dumps:
        from kubemarine import controlplane
        inventory_for_dump = controlplane.controlplane_finalize_inventory(cluster, prepare_for_dump(inventory))
        utils.dump_file(cluster, yaml.dump(inventory_for_dump, ), "cluster.yaml")

    return inventory


def compile_inventory(inventory: dict, cluster: KubernetesCluster) -> dict:

    # convert references in yaml to normal values
    iterations = 100
    root = deepcopy(inventory)
    root['globals'] = static.GLOBALS
    root['env'] = os.Environ()

    while iterations > 0:

        cluster.log.verbose('Inventory is not rendered yet...')
        inventory = compile_object(cluster.log, inventory, root)

        temp_dump = yaml.dump(inventory)

        # remove golang specific
        temp_dump = re.sub(escaped_expression_regex, '', temp_dump.replace('\n', ''))

        # it is necessary to carry out several iterations,
        # in case we have dynamic variables that reference each other
        if '{{' in temp_dump or '{%' in temp_dump:
            iterations -= 1
        else:
            iterations = 0

    inventory = compile_object(cluster.log, inventory, root, ignore_jinja_escapes=False)

    from kubemarine import controlplane
    inventory_for_dump = controlplane.controlplane_finalize_inventory(cluster, prepare_for_dump(inventory))
    merged_inventory = yaml.dump(inventory_for_dump)
    utils.dump_file(cluster, merged_inventory, "cluster_precompiled.yaml")

    return inventory


def compile_object(logger: log.EnhancedLogger, struct: Any, root: dict, ignore_jinja_escapes: bool = True) -> Any:
    if isinstance(struct, list):
        new_struct = []
        for i, v in enumerate(struct):
            struct[i] = compile_object(logger, v, root, ignore_jinja_escapes=ignore_jinja_escapes)
            # delete empty list entries, which can appear after jinja compilation
            if struct[i] != '':
                new_struct.append(struct[i])
        struct = new_struct
    elif isinstance(struct, dict):
        for k, v in struct.items():
            struct[k] = compile_object(logger, v, root, ignore_jinja_escapes=ignore_jinja_escapes)
    elif isinstance(struct, str) and ('{{' in struct or '{%' in struct):
        struct = compile_string(logger, struct, root, ignore_jinja_escapes=ignore_jinja_escapes)

    return struct


def compile_string(logger: log.EnhancedLogger, struct: str, root: dict,
                   ignore_jinja_escapes: bool = True) -> str:
    logger.verbose("Rendering \"%s\"" % struct)

    if ignore_jinja_escapes:
        iterator = escaped_expression_regex.finditer(struct)
        struct = re.sub(escaped_expression_regex, '', struct)
        struct = jinja.new(logger, True, root).from_string(struct).render(**root)

        # TODO this does not work for {raw}{jinja}{raw}{jinja}
        for match in iterator:
            span = match.span()
            struct = struct[:span[0]] + match.group() + struct[span[0]:]
    else:
        struct = jinja.new(logger, True, root).from_string(struct).render(**root)

    logger.verbose("\tRendered as \"%s\"" % struct)
    return struct


def escape_jinja_characters_for_inventory(cluster: KubernetesCluster, obj: Any) -> Any:
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = escape_jinja_characters_for_inventory(cluster, value)
    elif isinstance(obj, str):
        obj = _escape_jinja_character(obj)
    return obj


def _escape_jinja_character(value: str) -> str:
    if '{{' in value and '}}' in value and re.search(jinja_query_regex, value):
        matches = re.findall(jinja_query_regex, value)
        for match in matches:
            # TODO: rewrite to correct way of match replacement: now it can cause "{raw}{raw}xxx.." circular bug
            value = value.replace(match, '{% raw %}'+match+'{% endraw %}')
    return value


def prepare_for_dump(inventory: dict, copy: bool = True) -> dict:
    # different preparations before the inventory can be dumped

    if copy:
        dump_inventory = deepcopy(inventory)
    else:
        dump_inventory = inventory

    return dump_inventory


def manage_true_false_values(inventory: dict, _: KubernetesCluster) -> dict:
    # Check undefined values for plugin.name.install and convert it to bool
    for plugin_name, plugin_item in inventory["plugins"].items():
        # Check install value
        if 'install' not in plugin_item:
            continue
        value = utils.true_or_false(plugin_item.get('install', False))
        if value == 'undefined':
            raise ValueError(f"Found unsupported value for plugin.{plugin_name}.install: {plugin_item['install']}")
        plugin_item['install'] = value == 'true'
    return inventory
